# -*- coding: utf-8 -*-
"""
Log information about the number and rate of rendezvous connections to a HS.
"""

import sys
import time
import argparse
import logging
import logging.handlers
import threading

import stem
from stem.control import Controller
import schedule

handler = logging.StreamHandler()
formatter = logging.Formatter(fmt="%(asctime)s [%(levelname)s]: %(message)s")
handler.setFormatter(formatter)

logger = logging.getLogger("onionbalance")
logger.addHandler(handler)
logger.setLevel(logging.DEBUG)

lock = threading.RLock()

# Track circuits established in current time period.
new_rend_circuits_established = 0
rend_circuits_closed = 0


def circ_event_handler(event):
    """
    Handle the event received when Tor emits an event related the a rendezvous
    circuit
    """
    global new_rend_circuits_established, rend_circuits_closed

    if event.purpose == "HS_SERVICE_REND" and event.hs_state == "HSSR_JOINED":
        if event.type == "CIRC_MINOR":
            # Log when a new rendezvous circuit is successfully established.
            # A CIRC_MINOR event is emitted when the rendezvous circuit moves
            # from HS_STATE=HSSR_CONNECTING to HS_STATE=HSSR_JOINED
            logger.debug("New rendezvous circuit established (CircID: %s)",
                         event.id)
            new_rend_circuits_established += 1

        elif event.type == "CIRC" and event.status == "CLOSED":
            logger.debug("Rendezvous circuit closed (CircID: %s)", event.id)
            rend_circuits_closed += 1
    return


def output_status(controller):
    """
    Output the current counts every tick period.
    """
    global new_rend_circuits_established, rend_circuits_closed

    # Count number of currently established rendezvous circuits for this HS.
    rend_circ_count = len([circ for circ in controller.get_circuits()
                           if circ.purpose == "HS_SERVICE_REND"
                           and circ.hs_state == "HSSR_JOINED"])

    with lock:
        logger.info("New rend circuits: %d - Closed rend circuits: %d - "
                    "Established rend circuits: %d",
                    new_rend_circuits_established,
                    rend_circuits_closed,
                    rend_circ_count)
        new_rend_circuits_established = 0
        rend_circuits_closed = 0

    return None


def parse_cmd_args():
    """
    Parses and returns command line arguments.
    """

    parser = argparse.ArgumentParser(
        description="%s logs stats about Tor rendezvous circuits" %
        sys.argv[0])

    parser.add_argument("-i", "--ip", type=str, default="127.0.0.1",
                        help="Tor controller IP address")

    parser.add_argument("-p", "--port", type=int, default=9051,
                        help="Tor controller port")

    parser.add_argument("-t", "--tick", type=int, default=60,
                        help="Output total every tick seconds "
                        "(default: %(default)s)")

    parser.add_argument("--log-file", type=str, default="rendezvous.log",
                        help="Location to log the rendezvous connection"
                        "data.")

    parser.add_argument("-v", "--verbosity", type=str, default="info",
                        help="Minimum verbosity level for logging.  Available "
                             "in ascending order: debug, info, warning, "
                             "error, critical).  The default is info.")

    return parser.parse_args()


def main():

    args = parse_cmd_args()
    logger.setLevel(logging.__dict__[args.verbosity.upper()])

    if args.log_file:
        file_handler = logging.handlers.TimedRotatingFileHandler(
            args.log_file, when='D')
        file_handler.setFormatter(formatter)
        logger.addHandler(file_handler)

    logger.info("Beginning rendezvous circuit monitoring."
                "Status output every %d seconds", args.tick)

    with Controller.from_port(port=args.port) as controller:
        # Create a connection to the Tor control port
        controller.authenticate()

        # Add event listeners for HS_DESC and HS_DESC_CONTENT
        controller.add_event_listener(circ_event_handler,
                                      stem.control.EventType.CIRC)
        controller.add_event_listener(circ_event_handler,
                                      stem.control.EventType.CIRC_MINOR)

        # Schedule rendezvous status output.
        schedule.every(args.tick).seconds.do(output_status, controller)
        schedule.run_all()

        try:
            while True:
                schedule.run_pending()
                time.sleep(1)
        except KeyboardInterrupt:
            logger.info("Stopping rendezvous circuit monitoring.")

    sys.exit(0)

if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
from builtins import str, object

from onionbalance.common import log

from onionbalance.hs_v2 import descriptor
from onionbalance.hs_v2 import consensus

logger = log.get_logger()


class EventHandler(object):

    """
    Handles asynchronous Tor events.
    """

    @staticmethod
    def new_status(status_event):
        """
        Parse Tor status events such as "STATUS_CLIENT"
        """
        # pylint: disable=no-member
        if status_event.action == "CONSENSUS_ARRIVED":
            # Update the local view of the consensus in Onionbalance
            try:
                consensus.refresh_consensus()
            except Exception:
                logger.exception("An unexpected exception occured in the "
                                 "when processing the consensus update "
                                 "callback.")

    @staticmethod
    def new_desc(desc_event):
        """
        Parse HS_DESC response events
        """
        logger.debug("Received new HS_DESC event: %s", str(desc_event))

    @staticmethod
    def new_desc_content(desc_content_event):
        """
        Parse HS_DESC_CONTENT response events for descriptor content

        Update the HS instance object with the data from the new descriptor.
        """
        logger.debug("Received new HS_DESC_CONTENT event for %s.onion",
                     desc_content_event.address)

        #  Check that the HSDir returned a descriptor that is not empty
        descriptor_text = str(desc_content_event.descriptor).encode('utf-8')

        # HSDirs provide a HS_DESC_CONTENT response with either one or two
        # CRLF lines when they do not have a matching descriptor. Using
        # len() < 5 should ensure all empty HS_DESC_CONTENT events are matched.
        if len(descriptor_text) < 5:
            logger.debug("Empty descriptor received for %s.onion",
                         desc_content_event.address)
            return None

        # Send content to callback function which will process the descriptor
        try:
            descriptor.descriptor_received(descriptor_text)
        except Exception:
            logger.exception("An unexpected exception occured in the "
                             "new descriptor callback.")

        return None



import os
import unittest
# import mock
from datetime import datetime

import yaml

from onionbalance.hs_v3.manager import status_socket_location
from onionbalance.hs_v3.status import StatusSocketHandlerMixin


class DummyDescriptor(object):
    pass


class DummyInstance(object):
    pass


class DummyService(object):
    def __init__(self):
        self.instances = []


class TestStatus(unittest.TestCase):
    def test_status(self):
        # Mock a fake Tor network
        from onionbalance.hs_v3.onionbalance import my_onionbalance

        time_format = "%Y-%m-%d %H:%M:%S"

        service = DummyService()
        service.onion_address = 'bvy46sg2b5dokczabwv2pabqlrps3lppweyrebhat6gjieo2avojdvad.onion'
        service.first_descriptor = None
        service.second_descriptor = None
        my_onionbalance.services = [service]
        status = StatusSocketHandlerMixin(my_onionbalance)
        self.assertEqual(status._outputString(), '{"services": [{"instances": [], "onionAddress": "bvy46sg2b5dokczabwv2pabqlrps3lppweyrebhat6gjieo2avojdvad.onion", "publishAttemptFirstDescriptor": null, "publishAttemptSecondDescriptor": null}]}')

        service.first_descriptor = DummyDescriptor()
        service.first_descriptor.last_publish_attempt_ts = datetime.strptime('2020-06-16 20:00:12', time_format)
        service.first_descriptor.last_upload_ts = datetime.strptime('2020-06-16 20:00:13', time_format)
        service.second_descriptor = DummyDescriptor()
        service.second_descriptor.last_publish_attempt_ts = datetime.strptime('2020-06-16 20:00:14', time_format)
        service.second_descriptor.last_upload_ts = datetime.strptime('2020-06-16 20:00:15', time_format)
        self.assertEqual(status._outputString(), '{"services": [{"instances": [], "onionAddress": "bvy46sg2b5dokczabwv2pabqlrps3lppweyrebhat6gjieo2avojdvad.onion", "publishAttemptFirstDescriptor": "2020-06-16 20:00:12", "publishAttemptSecondDescriptor": "2020-06-16 20:00:14"}]}')

        service.first_descriptor = None
        service.second_descriptor = None
        instance1 = DummyInstance()
        instance1.onion_address = 'vkmiy6biqcyphtx5exswxl5sjus2vn2b6pzir7lz5akudhwbqk5muead'
        instance1.intro_set_modified_timestamp = datetime.strptime('2020-06-16 21:00:12', time_format)
        instance1.descriptor = DummyDescriptor()
        instance1.descriptor.received_ts = datetime.strptime('2020-06-11 20:00:10', time_format)
        instance1.descriptor.intro_set = ['127.0.0.1', '127.0.0.2']
        instance2 = DummyInstance()
        instance2.onion_address = 'jhkhjkhdgjkhdfjkhgfjkhgkjdhgjkfdhgjkfdhhdfgjfhdgkjfhkd88'
        instance2.intro_set_modified_timestamp = None
        instance2.descriptor = DummyDescriptor()
        instance2.descriptor.received_ts = datetime.strptime('2020-07-11 20:00:10', time_format)
        instance2.descriptor.intro_set = ['127.0.0.1', '127.0.0.2']
        service.instances = [instance1, instance2]
        self.assertEqual(status._outputString(), '{"services": [{"instances": [{"descriptorReceived": "2020-06-11 20:00:10", "introPointsNum": 2, "introSetModified": "2020-06-16 21:00:12", "onionAddress": "vkmiy6biqcyphtx5exswxl5sjus2vn2b6pzir7lz5akudhwbqk5muead.onion"}, {"onionAddress": "jhkhjkhdgjkhdfjkhgfjkhgkjdhgjkfdhgjkfdhhdfgjfhdgkjfhkd88.onion", "received": "2020-07-11 20:00:10"}], "onionAddress": "bvy46sg2b5dokczabwv2pabqlrps3lppweyrebhat6gjieo2avojdvad.onion", "publishAttemptFirstDescriptor": null, "publishAttemptSecondDescriptor": null}]}')


class TestStatusSocketLocation(unittest.TestCase):
    def test_location(self):
        try:
            del os.environ['ONIONBALANCE_STATUS_SOCKET_LOCATION']
        except KeyError:
            pass
        config_data = yaml.safe_load("dummy: x")
        self.assertEqual(status_socket_location(config_data), None)
        config_data = yaml.safe_load("status-socket-location: /home/user/test.sock")
        self.assertEqual(status_socket_location(config_data), '/home/user/test.sock')
        os.environ['ONIONBALANCE_STATUS_SOCKET_LOCATION'] = '/home/user/test2.sock'
        self.assertEqual(status_socket_location(config_data), '/home/user/test2.sock')


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
import datetime
import time
import base64

import Cryptodome.PublicKey.RSA
import stem

from onionbalance.hs_v2 import descriptor
from onionbalance.hs_v2 import util
from onionbalance.hs_v2 import consensus

import onionbalance.common.descriptor
import onionbalance.common.util

from onionbalance.common import log
from onionbalance.hs_v2 import config
from onionbalance.common import intro_point_set

logger = log.get_logger()


def publish_all_descriptors():
    """
    Called periodically to upload new super-descriptors if needed

    .. todo:: Publishing descriptors for different services at the same time
              will leak that they are related. Descriptors should
              be published individually at a random interval to avoid
              correlation.
    """
    logger.debug("Checking if any master descriptors should be published.")
    for service in config.services:
        service.descriptor_publish()


class Service(object):
    """
    Service represents a front-facing hidden service which should
    be load-balanced.
    """

    def __init__(self, controller, service_key=None, instances=None):
        """
        Initialise a HiddenService object.
        """
        self.controller = controller

        # Service key must be a valid PyCrypto RSA key object
        if isinstance(service_key, Cryptodome.PublicKey.RSA.RsaKey):
            self.service_key = service_key
        else:
            raise ValueError("Service key is not a valid RSA object.")

        # List of instances for this onion service
        if not instances:
            instances = []
        self.instances = instances

        # Calculate the onion address for this service
        self.onion_address = util.calc_onion_address(self.service_key)

        # Timestamp when this descriptor was last attempted
        self.uploaded = None

    def _intro_points_modified(self):
        """
        Check if the introduction point set has changed since last
        publish.
        """
        return any(instance.intro_set_changed_since_published
                   for instance in self.instances)

    def _descriptor_not_uploaded_recently(self):
        """
        Check if the master descriptor hasn't been uploaded recently
        """
        if not self.uploaded:
            # Descriptor never uploaded
            return True

        descriptor_age = (datetime.datetime.utcnow() - self.uploaded)
        if (descriptor_age.total_seconds() > config.DESCRIPTOR_UPLOAD_PERIOD):
            return True
        else:
            return False

    def _descriptor_id_changing_soon(self):
        """
        If the descriptor ID will change soon, upload under both descriptor IDs
        """
        permanent_id = base64.b32decode(self.onion_address, 1)
        seconds_valid = util.get_seconds_valid(time.time(), permanent_id)

        # Check if descriptor ID will be changing within the overlap period.
        if seconds_valid < config.DESCRIPTOR_OVERLAP_PERIOD:
            return True
        else:
            return False

    def _select_introduction_points(self):
        """
        Choose set of introduction points from all fresh descriptors

        Returns an intro_point_set.IntroductionPointSet() which can be used to
        choose introduction points.
        """
        available_intro_points = []

        # Loop through each instance and determine fresh intro points
        for instance in self.instances:
            if not instance.received:
                logger.info("No descriptor received for instance %s.onion "
                            "yet.", instance.onion_address)
                continue

            # The instance may be offline if no descriptor has been received
            # for it recently or if the received descriptor's timestamp is
            # too old
            received_age = datetime.datetime.utcnow() - instance.received
            timestamp_age = datetime.datetime.utcnow() - instance.timestamp
            received_age = received_age.total_seconds()
            timestamp_age = timestamp_age.total_seconds()

            if received_age > config.DESCRIPTOR_UPLOAD_PERIOD:
                logger.info("Our descriptor for instance %s.onion "
                            "was received too long ago (%d). "
                            "The instance may be offline. Its introduction "
                            "points will not be included in the master "
                            "descriptor.",
                            instance.onion_address,
                            received_age)
                continue
            elif timestamp_age > (4 * 60 * 60):
                logger.info("Our descriptor for instance %s.onion "
                            "has an old timestamp (%d). "
                            "The instance may be offline. Its introduction "
                            "points will not be included in the master "
                            "descriptor.",
                            instance.onion_address,
                            timestamp_age)
                continue
            else:
                # Include this instance's introduction points
                instance.intro_set_changed_since_published = False
                available_intro_points.append(instance.introduction_points)

        return intro_point_set.IntroductionPointSet(available_intro_points)

    def _publish_descriptor(self, deviation=0):
        """
        Create, sign and upload master descriptors for this service
        """

        # Retrieve the set of available introduction points
        intro_point_set = self._select_introduction_points()
        max_intro_points = config.MAX_INTRO_POINTS

        # Upload multiple unique descriptors which contain different
        # subsets of the available introduction points.
        # (https://github.com/DonnchaC/onionbalance/issues/16)
        distinct_descriptors = config.DISTINCT_DESCRIPTORS

        # If we have <= MAX_INTRO_POINTS we should choose the introduction
        # points now and use the same set in every descriptor. Using the
        # same set of introduction points will look more like a standard
        # Tor client.
        num_intro_points = len(intro_point_set)

        if num_intro_points <= max_intro_points:
            intro_points = intro_point_set.choose(num_intro_points)
            logger.debug("We have %d IPs, not using distinct descriptors.",
                         len(intro_point_set))
            distinct_descriptors = False

        for replica in range(0, config.REPLICAS):
            # Using distinct descriptors, choose a new set of intro points
            # for each descriptor and upload it to individual HSDirs.
            if distinct_descriptors:
                descriptor_id = util.calc_descriptor_id_b32(
                    self.onion_address,
                    time=time.time(),
                    replica=replica,
                    deviation=deviation,
                )
                responsible_hsdirs = consensus.get_hsdirs(descriptor_id)

                for hsdir in responsible_hsdirs:
                    intro_points = intro_point_set.choose(max_intro_points)
                    try:
                        signed_descriptor = (
                            descriptor.generate_service_descriptor(
                                self.service_key,
                                introduction_point_list=intro_points,
                                replica=replica,
                                deviation=deviation
                            ))
                    except ValueError as exc:
                        logger.warning("Error generating descriptor: %s", exc)
                        continue

                    # Signed descriptor was generated successfully, upload it
                    # to the respective HSDir
                    self._upload_descriptor(signed_descriptor, replica,
                                            hsdirs=hsdir)
                logger.info("Published distinct master descriptors for "
                            "service %s.onion under replica %d.",
                            self.onion_address, replica)

            else:
                # Not using distinct descriptors, upload one descriptor
                # under each replica and let Tor pick the HSDirs.
                try:
                    signed_descriptor = descriptor.generate_service_descriptor(
                        self.service_key,
                        introduction_point_list=intro_points,
                        replica=replica,
                        deviation=deviation
                    )
                except ValueError as exc:
                    logger.warning("Error generating descriptor: %s", exc)
                    continue

                # Signed descriptor was generated successfully, upload it
                self._upload_descriptor(signed_descriptor, replica)
                logger.info("Published a descriptor for service %s.onion "
                            "under replica %d.", self.onion_address, replica)

        # It would be better to set last_uploaded when an upload succeeds and
        # not when an upload is just attempted. Unfortunately the HS_DESC #
        # UPLOADED event does not provide information about the service and
        # so it can't be used to determine when descriptor upload succeeds
        self.uploaded = datetime.datetime.utcnow()

    def _upload_descriptor(self, signed_descriptor, replica, hsdirs=None):
        """
        Convenience method to upload a descriptor
        Handle some error checking and logging inside the Service class
        """
        if hsdirs and not isinstance(hsdirs, list):
            hsdirs = [hsdirs]

        while True:
            try:
                onionbalance.common.descriptor.upload_descriptor(self.controller,
                                                                 signed_descriptor,
                                                                 hsdirs=hsdirs)
                break
            except stem.SocketClosed:
                logger.error("Error uploading descriptor for service "
                             "%s.onion, Socket is closed.",
                             self.onion_address)
                onionbalance.common.util.reauthenticate(self.controller, logger)
            except stem.ControllerError:
                logger.exception("Error uploading descriptor for service "
                                 "%s.onion.", self.onion_address)
                break

    def descriptor_publish(self, force_publish=False):
        """
        Publish descriptor if have new IPs or if descriptor has expired
        """

        # A descriptor should be published if any of the following conditions
        # are True
        if any([self._intro_points_modified(),  # If any IPs have changed
                self._descriptor_not_uploaded_recently(),
                force_publish]):

            logger.debug("Publishing a descriptor for service %s.onion.",
                         self.onion_address)
            self._publish_descriptor()

            # If the descriptor ID will change soon, need to upload under
            # the new ID too.
            if self._descriptor_id_changing_soon():
                logger.info("Publishing a descriptor for service %s.onion "
                            "under next descriptor ID.", self.onion_address)
                self._publish_descriptor(deviation=1)

        else:
            logger.debug("Not publishing a new descriptor for service "
                         "%s.onion.", self.onion_address)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# onionbalance documentation build configuration file, created by
# sphinx-quickstart on Wed Jun 10 13:54:42 2015.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

import sys
import os
import datetime

import sphinx.environment
from docutils.utils import get_source_line

# Documentation configuration
__version__ = '0.2.2'
__author__ = "George Kadianakis, Donncha O'Cearbhaill"
__contact__ = "asn@torproject.org"

# Ignore the 'dev' version suffix.
if __version__.endswith('dev'):
    __version__ = __version__[:-4]


# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
sys.path.insert(0, os.path.abspath('..'))

on_rtd = os.environ.get('READTHEDOCS', None) == 'True'

# -- General configuration ------------------------------------------------


# Don't give warning for external images
def _warn_node(self, msg, node):
    if not msg.startswith('nonlocal image URI found:'):
        self._warnfunc(msg, '%s:%s' % get_source_line(node))
sphinx.environment.BuildEnvironment.warn_node = _warn_node

# If your documentation needs a minimal Sphinx version, state it here.
needs_sphinx = '1.1'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'alabaster',
    'sphinx.ext.autodoc',
    'sphinx.ext.todo',
    'sphinx.ext.viewcode',
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
# source_suffix = ['.rst', '.md']
source_suffix = '.rst'

# The encoding of source files.
source_encoding = 'utf-8-sig'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = 'Onionbalance'

# Remove copyright notice for man page
copyright = ''
author = __author__

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = __version__
# The full version, including alpha/beta/rc tags.
release = __version__

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = 'en'

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
#today = ''
# Else, today_fmt is used as the format for a strftime call.
#today_fmt = '%B %d, %Y'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ['_build', 'modules.rst']

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# A list of ignored prefixes for module index sorting.
#modindex_common_prefix = []

# If true, keep warnings as "system message" paragraphs in the built documents.
#keep_warnings = False

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = True


# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = 'alabaster'

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
html_theme_options = {
    "description": "Load balancing and redundancy for Tor onion services.",
    'github_user': 'asn-d6',
    'github_repo': 'onionbalance',
    'github_button': False,
    'travis_button': False,
}

# Enable external resources on the RTD hosted documentation only
if on_rtd:
    html_theme_options['github_button'] = True
    html_theme_options['travis_button'] = True

# Add any paths that contain custom themes here, relative to this directory.
#html_theme_path = []

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
#html_title = None

# A shorter title for the navigation bar.  Default is the same as html_title.
html_short_title = "Onionbalance Docs"

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
#html_logo = None

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = []

# Custom sidebar templates, maps document names to template names.
html_sidebars = {
    '**': [
        'about.html',
        'navigation.html',
        'relations.html',
    ]
}

# If false, no module index is generated.
html_domain_indices = False

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
html_show_sphinx = False

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
html_show_copyright = False

# Output file base name for HTML help builder.
htmlhelp_basename = 'onionbalancedoc'

# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    ('running-onionbalance', 'onionbalance',
      'a Tor onion service load balancer',
     ['%s <%s>' % (__author__, __contact__)], 1),
    ('onionbalance-config', 'onionbalance-config',
      'tool for generating onionbalance config files and keys',
     ['%s <%s>' % (__author__, __contact__)], 1),
]

# If true, show URL addresses after external links.
#man_show_urls = False



from stem.util import ed25519

from cryptography.hazmat.primitives.asymmetric.ed25519 import Ed25519PublicKey, Ed25519PrivateKey
from cryptography.hazmat.primitives import serialization

from onionbalance.hs_v3.ext import ed25519_exts_ref
from onionbalance.hs_v3.ext import slow_ed25519

def load_tor_key_from_disk(key_bytes):
    """
    Load a private identity key from little-t-tor.
    """
    # Verify header
    if (key_bytes[:29] != b'== ed25519v1-secret: type0 =='):
        raise ValueError("Tor key does not start with Tor header")

    expanded_sk = key_bytes[32:]

    # The rest should be 64 bytes (a,h):
    # 32 bytes for secret scalar 'a'
    # 32 bytes for PRF key 'h'
    if (len(expanded_sk) != 64):
        raise ValueError("Tor private key has the wrong length")

    return TorEd25519PrivateKey(expanded_sk)

def _blinded_sign_with_tor_key(msg, identity_key, blinded_key, blinding_nonce):
    """
    This is identical to stem's hidden_service.py:_blinded_sign() but takes an
    extended private key (i.e. in tor format) as its argument, instead of the
    standard format that hazmat does. It basically omits the "extended the key"
    step and does everything else the same.
    """
    identity_key_bytes = identity_key.private_bytes(
        encoding = serialization.Encoding.Raw,
        format = serialization.PrivateFormat.Raw,
        encryption_algorithm = serialization.NoEncryption(),
    )

    # blind the ESK with this nonce
    esk = identity_key_bytes

    mult = 2 ** (ed25519.b - 2) + sum(2 ** i * ed25519.bit(blinding_nonce, i) for i in range(3, ed25519.b - 2))
    s = ed25519.decodeint(esk[:32])
    s_prime = (s * mult) % ed25519.l
    k = esk[32:]
    k_prime = ed25519.H(b'Derive temporary signing key hash input' + k)[:32]
    blinded_esk = ed25519.encodeint(s_prime) + k_prime

    # finally, sign the message

    a = ed25519.decodeint(blinded_esk[:32])
    r = ed25519.Hint(b''.join([blinded_esk[i:i + 1] for i in range(ed25519.b // 8, ed25519.b // 4)]) + msg)
    R = ed25519.scalarmult(ed25519.B, r)
    S = (r + ed25519.Hint(ed25519.encodepoint(R) + blinded_key + msg) * a) % ed25519.l

    return ed25519.encodepoint(R) + ed25519.encodeint(S)

"""
Tor ed25519 keys

Expose classes for ed25519 keys in Tor's extended key format which can mimic
the hazmat ed25519 public/private key classes, so that we can use them
interchangeably in stem.

Tor uses the "extended" (a,h) format for its private keys, whereas hazmat uses
the "standard" (seed,A) format: https://blog.mozilla.org/warner/2011/11/29/ed25519-keys/

Since you can't go from the "extended" format to the "standard" format, we
created these wrappers that act exactly like hazmat keys when it comes to their
interface.
"""


class TorEd25519PrivateKey(object):
    """
    Represents the private part of a blinded ed25519 key of an onion service
    and should expose a public_key() method and a sign() method.
    """
    def __init__(self, expanded_sk):
        self.priv_key = expanded_sk
        self.pub_key_bytes = ed25519_exts_ref.publickeyFromESK(self.priv_key)
        self.pub_key = TorEd25519PublicKey(self.pub_key_bytes)

    def public_key(self):
        return self.pub_key

    def private_bytes(self, encoding=None, format=None, encryption_algorithm=None):
        return self.priv_key

    def sign(self, msg):
        return ed25519_exts_ref.signatureWithESK(msg, self.priv_key, self.pub_key_bytes)

    @property
    def __class__(self):
        """
        This is an epic hack to make this class look like a hazmat ed25519 public
        key in the eyes of stem:
          https://github.com/asn-d6/onionbalance/issues/10#issuecomment-610425916

        The __class_ attribute is what's being used by unittest.mock and the C
        API to trick isinstance() checks, so as long as stem uses isinstance()
        this is gonna work:
          https://docs.python.org/3/library/unittest.mock.html#unittest.mock.Mock.__class__
          https://docs.python.org/3/c-api/object.html#c.PyObject_IsInstance
        """
        return Ed25519PrivateKey

class TorEd25519PublicKey(object):
    """
    Represents the public blinded ed25519 key of an onion service and should
    expose a public_bytes() method and a verify() method.
    """
    def __init__(self, public_key):
        self.public_key = public_key

    def public_bytes(self, encoding=None, format=None):
        return self.public_key

    def verify(self, signature, message):
        """
        raises exception if sig not valid
        """
        slow_ed25519.checkvalid(signature, message, self.public_key)

    @property
    def __class__(self):
        return Ed25519PublicKey



# -*- coding: utf-8 -*-
import pytest

from onionbalance.hs_v2 import consensus
from onionbalance.hs_v2 import config

# Mock hex-encoded HSDir fingerprint list
MOCK_HSDIR_LIST = [
    "1111111111111111111111111111111111111111",
    "2222222222222222222222222222222222222222",
    "3333333333333333333333333333333333333333",
    "4444444444444444444444444444444444444444",
    "5555555555555555555555555555555555555555",
    "6666666666666666666666666666666666666666",
]

config.HSDIR_SET = 3  # Always select 3 responsible HSDirs


def test_get_hsdirs_no_consensus():
    """
    `get_hsdirs` should raise an exception when we don't have a valid
    HSDir list from the consensus.
    """

    with pytest.raises(ValueError):
        consensus.get_hsdirs('aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa')


def test_get_hsdirs(monkeypatch):
    """Test for normal responsible HSDir selection"""

    monkeypatch.setattr(consensus, 'HSDIR_LIST', MOCK_HSDIR_LIST)

    # Descriptor ID before '222....''
    descriptor_id_base32 = "eiqaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"
    responsible_hsdirs = consensus.get_hsdirs(descriptor_id_base32)

    assert (responsible_hsdirs == [
        "2222222222222222222222222222222222222222",
        "3333333333333333333333333333333333333333",
        "4444444444444444444444444444444444444444",
    ])


def test_get_hsdirs_edge_of_ring(monkeypatch):
    """Test that selection wraps around the edge of the HSDir ring"""

    monkeypatch.setattr(consensus, 'HSDIR_LIST', MOCK_HSDIR_LIST)

    # Descriptor ID before '666....''
    descriptor_id_base32 = "mzqaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"
    responsible_hsdirs = consensus.get_hsdirs(descriptor_id_base32)

    assert (responsible_hsdirs == [
        "6666666666666666666666666666666666666666",
        "1111111111111111111111111111111111111111",
        "2222222222222222222222222222222222222222",
    ])


def test_get_hsdirs_no_repeat(monkeypatch):
    """Test that selection wraps around the edge of the HSDir ring"""

    SHORT_HSDIR_LIST = [
        "1111111111111111111111111111111111111111",
        "2222222222222222222222222222222222222222",
    ]
    monkeypatch.setattr(consensus, 'HSDIR_LIST', SHORT_HSDIR_LIST)

    # Descriptor ID before '111....''
    descriptor_id_base32 = "ceiaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"
    responsible_hsdirs = consensus.get_hsdirs(descriptor_id_base32)

    assert (responsible_hsdirs == [
        "1111111111111111111111111111111111111111",
        "2222222222222222222222222222222222222222",
    ])



import stem

from onionbalance.common import log

logger = log.get_logger()


def upload_descriptor(controller, signed_descriptor, hsdirs=None, v3_onion_address=None):
    """
    Upload descriptor via the Tor control port

    If no HSDirs are specified, Tor will upload to what it thinks are the
    responsible directories

    If 'v3_onion_address' is set, this is a v3 HSPOST request, and the address
    needs to be embedded in the request.
    """
    logger.debug("Beginning service descriptor upload.")

    server_args = ""

    # Provide server fingerprints to control command if HSDirs are specified.
    if hsdirs:
        server_args = ' '.join([("SERVER={}".format(hsdir))
                                for hsdir in hsdirs])

    if v3_onion_address:
        server_args += " HSADDRESS=%s" % v3_onion_address.replace(".onion", "")

    # Stem will insert the leading + and trailing '\r\n.\r\n'
    response = controller.msg("HSPOST %s\n%s" %
                              (server_args, signed_descriptor))

    (response_code, divider, response_content) = response.content()[0]
    if not response.is_ok():
        if response_code == "552":
            raise stem.InvalidRequest(response_code, response_content)
        else:
            raise stem.ProtocolError("HSPOST returned unexpected response "
                                     "code: %s\n%s" % (response_code,
                                                       response_content))



from onionbalance.hs_v3 import consensus

def test_disaster_srv():
    """
    Test that disaster SRV creation is correct based on little-t-tor's
    unittests as test vectors (in particular see test_disaster_srv())
    """
    my_consensus = consensus.Consensus(do_refresh_consensus=False)

    # Correct disaster SRV for TPs: 1, 2, 3, 4, 5
    correct_srvs = [ "F8A4948707653837FA44ABB5BBC75A12F6F101E7F8FAF699B9715F4965D3507D",
                     "C17966DF8B4834638E1B7BF38944C4995B92E89749D1623E417A44938D08FD67",
                     "A3BAB4327F9C2F8B30A126DAD3ABCCE12DD813169A5D924244B57987AEE413C2",
                     "C79ABE7116FCF7810F7A5C76198B97339990C738B456EFDBC1399189927BADEC",
                     "941D5FE4289FBF2F853766EAC4B948B51C81A4137A44516342ABC80518E0183D"]

    for i in range(1,6):
        disaster_srv = my_consensus._get_disaster_srv(i)
        assert(disaster_srv.hex().upper() == correct_srvs[i-1])

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
"""
Simple scheduler for running jobs at regular intervals
"""
import functools
import time

from onionbalance.common import log

logger = log.get_logger()

jobs = []


class Job(object):
    """
    Object to represent a scheduled job task
    """

    def __init__(self, interval, job_func, *job_args, **job_kwargs):
        self.interval = interval
        self.planned_run_time = time.time()

        # Configure the job function and calling arguments
        self.job_func = functools.partial(job_func, *job_args, **job_kwargs)
        functools.update_wrapper(self.job_func, job_func)

    def __lt__(self, other):
        """
        Jobs are sorted based on their next scheduled run time
        """
        return self.planned_run_time < other.planned_run_time

    @property
    def should_run(self):
        """
        Check if the job should be run now
        """
        return self.planned_run_time <= time.time()

    def run(self, override_run_time=None):
        """
        Run job then reschedule it in the job list
        """
        ret = self.job_func()

        # Pretend the job was scheduled now, if we ran it early with run_all()
        if override_run_time:
            self.planned_run_time = time.time()
        self.planned_run_time += self.interval

        return ret

    def __repr__(self):
        """
        Return human readable representation of the Job and arguments
        """
        args = [repr(x) for x in self.job_func.args]
        kwargs = ["{}={}".format(k, repr(v)) for
                  k, v in self.job_func.keywords.items()]
        return "{}({})".format(self.job_func.__name__,
                               ', '.join(args + kwargs))


def add_job(interval, function, *job_args, **job_kwargs):
    """
    Add a job to be executed at regular intervals

    The `interval` value is in seconds, starting from now.
    """
    job = Job(interval, function, *job_args, **job_kwargs)
    jobs.append(job)


def _run_job(job, override_run_time=False):
    """
    Run a job and put it back in the job queue
    """
    return job.run(override_run_time)


def run_all(delay_seconds=0):
    """
    Run all jobs at `delay_seconds` regardless of their schedule
    """
    for job in jobs:
        _run_job(job, override_run_time=True)
        time.sleep(delay_seconds)


def run_forever(check_interval=1):
    """
    Run jobs forever
    """
    while True:
        if not jobs:
            logger.error("No scheduled jobs found, scheduler exiting.")
            return None

        jobs_to_run = (job for job in jobs if job.should_run)
        for job in sorted(jobs_to_run):
            _run_job(job)

        time.sleep(check_interval)



# -*- coding: utf-8 -*-
import signal
import sys
import time

import pexpect

from test.functional.util import *

def test_sighup_reload_config(tmpdir):
    """
    Functional test to run Onionbalance, send SIGHUP then check if config is reloaded
    """

    chutney_config = parse_chutney_enviroment()

    original_instance_address = random_onionv3_address()
    config_file_path = create_test_config_file_v3(tmppath=tmpdir, instance_address=original_instance_address)
    assert config_file_path

    # Start an Onionbalance server and monitor for correct output with pexpect
    server = pexpect.spawnu("onionbalance",
                            args=[
                                '--hs-version', 'v3',
                                '-i', chutney_config.get('client_ip'),
                                '-p', str(chutney_config.get('control_port')),
                                '-c', config_file_path,
                                '-v', 'debug',
                                '--is-testnet'
                            ], logfile=sys.stdout, timeout=5)
    time.sleep(1)

    # Check for expected output from Onionbalance
    server.expect(u"Loaded the config file")
    server.expect(original_instance_address)

    # Update config file and send SIGHUP
    updated_instance_address = random_onionv3_address()
    config_file_path = update_test_config_file_v3(tmppath=tmpdir, instance_address=updated_instance_address)
    assert config_file_path
    server.kill(signal.SIGHUP)
    server.expect(u"Signal SIGHUP received, reloading configuration")
    server.expect(u"Loaded the config file")
    server.expect(updated_instance_address)



# -*- coding: utf-8 -*-
"""
Functional tests which run the onionbalance-config tool and check
the created output files.
"""
import sys

import pexpect
import Cryptodome.PublicKey.RSA

import onionbalance.hs_v2.util


def onionbalance_config_interact(cli, cli_input):
    """
    Send each input line to the onionbalance-config CLI interface
    """
    cli.expect(u"Enter HS version")
    cli.send("v2\n")

    cli.expect(u"store generated config")
    cli.send("{}\n".format(cli_input.get('config_dir', u'')))

    cli.expect(u"path to master service private key")
    cli.send(u"{}\n".format(cli_input.get('private_key_path', u'')))

    cli.expect(u"Number of instance services")
    cli.send(u"{}\n".format(cli_input.get('num_instances', u'')))

    cli.expect(u"Provide a tag name")
    cli.send(u"{}\n".format(cli_input.get('tag_name', u'')))

    cli.expect(u"service virtual port")
    cli.send(u"{}\n".format(cli_input.get('virtual_port', u'')))

    cli.expect(u"service target IP and port")
    cli.send(u"{}\n".format(cli_input.get('target_ip', u'')))

    cli.expect(u"optional password")
    cli.send(u"{}\n".format(cli_input.get('password', u'')))

    return None


def check_basic_config_output(config_dir):
    """
    Run basic tests on the generated config files and keys to check
    that they look reasonable.
    """

    assert len(config_dir.listdir()) == 1 + 2

    # Find generated instance addresses
    instance_addresses = []
    for directory in config_dir.listdir():
        if directory.basename != 'master':
            instance_addresses.extend(
                [str(name.basename) for name in directory.listdir()
                 if 'torrc' not in name.basename])

    # Correct number of directories created
    assert len(config_dir.listdir()) == 1 + 2

    assert config_dir.join('master', 'torrc-server').check()
    assert config_dir.join('master', 'config.yaml').check()

    config_file = config_dir.join('master', 'config.yaml').read_text('utf-8')
    assert all(address in config_file for address in instance_addresses)

    # Test that all addresses are encoded as bytes and not unicode
    assert "!!python/unicode" not in config_file

    return True


def test_onionbalance_config_interactive(tmpdir):
    """
    Functional test to run onion-balance config in interactive mode.
    """
    # Start onionbalance-config in interactive mode (no command line arguments)
    cli = pexpect.spawnu("onionbalance-config", logfile=sys.stdout)
    cli.expect(u"entering interactive mode")

    # Interact with the running onionbalance-config process
    onionbalance_config_interact(
        cli, cli_input={'config_dir': str(tmpdir.join(u"configdir"))})
    cli.expect(u"Done! Successfully generated")

    check_basic_config_output(tmpdir.join(u"configdir"))


def test_onionbalance_config_automatic(tmpdir):
    """
    Functional test to run onion-balance config in automatic mode.
    """
    # Start onionbalance-config in automatic mode
    cli = pexpect.spawnu("onionbalance-config", logfile=sys.stdout,
                         args=[
                             '--hs-version', 'v2',
                             '--output', str(tmpdir.join(u"configdir")),
                         ])
    cli.expect(u"Done! Successfully generated")

    check_basic_config_output(tmpdir.join(u"configdir"))


def test_onionbalance_config_automatic_custom_ports(tmpdir):
    """
    Run onionbalance-config in interactive mode, providing a custom port line.
    """
    cli = pexpect.spawnu("onionbalance-config", logfile=sys.stdout,
                         args=[
                             '--hs-version', 'v2',
                             '--output', str(tmpdir.join(u"configdir")),
                             '--service-virtual-port', u'443',
                             '--service-target', u'127.0.0.1:8443',
                         ])
    cli.expect(u"Done! Successfully generated")

    # Read one of the generated torrc files
    for directory in tmpdir.join(u"configdir").listdir():
        if directory.basename != 'master':
            torrc_file = [name for name in directory.listdir()
                          if name.basename == 'instance_torrc'][0]
            break
    assert torrc_file.check()

    # Check torrc line contains the correct HiddenServicePort line
    torrc_contents = torrc_file.read_text('utf-8')
    assert u'HiddenServicePort 443 127.0.0.1:8443' in torrc_contents


def test_onionbalance_config_automatic_key_with_password(tmpdir, mocker):
    """
    Run onionbalance-config with an existing key, export as password protected
    key.
    """

    # Create input private_key
    private_key = Cryptodome.PublicKey.RSA.generate(1024)
    key_path = tmpdir.join('private_key')
    key_path.write(private_key.exportKey())

    # Start onionbalance-config in automatic mode
    cli = pexpect.spawnu("onionbalance-config", logfile=sys.stdout,
                         args=[
                             '--hs-version', 'v2',
                             '--output', str(tmpdir.join(u"configdir")),
                             '--key', str(key_path),
                             '--password', 'testpassword',
                         ])
    cli.expect(u"Done! Successfully generated")

    # Check master config was generated with password protected key.
    master_dir = tmpdir.join('configdir', 'master')
    output_key_path = [fpath for fpath in master_dir.listdir()
                       if fpath.ext == '.key'][0]
    assert output_key_path.check()

    # Check key decrypts and is valid
    mocker.patch('getpass.getpass', lambda *_: 'testpassword')
    output_key = onionbalance.hs_v2.util.key_decrypt_prompt(str(output_key_path))
    assert isinstance(output_key, Cryptodome.PublicKey.RSA.RsaKey)



import unittest
import binascii

from cryptography.hazmat.primitives.asymmetric.ed25519 import Ed25519PrivateKey, Ed25519PublicKey

import stem.util
from stem.descriptor.hidden_service import HiddenServiceDescriptorV3

from onionbalance.hs_v3 import tor_ed25519

ONION_ADDR = "4a2zwxetkje5lwvahfv75arzrctn6bznkwmosvfyobmyv2fc3idbpwyd.onion"
PRIVKEY_FILE_HEX = "3d3d206564323535313976312d7365637265743a207479706530203d3d000000a0c632db831865c9562c5eca9446c50d1c6a88d9985e372986c9ae800a7f347a93bd4a97b9b18adf2ad9bceee4d167bf3d61362440066c22719d3cb5d1f97011"

class TestKeys(unittest.TestCase):
    def test_load_tor_privkey(self):
        privkey_bytes = binascii.unhexlify(PRIVKEY_FILE_HEX)
        privkey = tor_ed25519.load_tor_key_from_disk(privkey_bytes)
        pubkey = privkey.public_key()

        # Make sure that the fake instances are right
        self.assertTrue(isinstance(privkey, Ed25519PrivateKey))
        self.assertTrue(isinstance(pubkey, Ed25519PublicKey))

        # Make sure that the public key matches the onion address
        onion_addr_pubkey_bytes = HiddenServiceDescriptorV3.identity_key_from_address(ONION_ADDR)
        self.assertEqual(onion_addr_pubkey_bytes, pubkey.public_bytes())

        # Check that signature verification works
        msg = b"07-04-2020 weird days"
        msg_sig = privkey.sign(msg)
        onion_addr_pubkey = tor_ed25519.TorEd25519PublicKey(onion_addr_pubkey_bytes)
        onion_addr_pubkey.verify(msg_sig, msg)

        # Now check that it won't just verify any message
        self.assertRaises(Exception, onion_addr_pubkey.verify, msg_sig, b"another message another day")

        # Now check that stem will accept this
        self.assertEqual(stem.util._pubkey_bytes(privkey), pubkey.public_bytes())
        self.assertEqual(stem.util._pubkey_bytes(pubkey), pubkey.public_bytes())

        # Now check that blinding can work
        blinded_key_bytes = stem.descriptor.hidden_service._blinded_pubkey(privkey, b"a"*32)
        blinded_key = tor_ed25519.TorEd25519PublicKey(blinded_key_bytes)
        signature = tor_ed25519._blinded_sign_with_tor_key(b"haha", privkey, blinded_key_bytes, b"a"*32)

        blinded_key.verify(signature, b"haha")

if __name__ == '__main__':
    unittest.main()




import datetime
import hashlib
import itertools

import stem.util
from stem.descriptor.hidden_service import HiddenServiceDescriptorV3, InnerLayer
from stem.descriptor.certificate import Ed25519CertificateV1, Ed25519Extension, ExtensionType

from cryptography.hazmat.primitives.asymmetric.ed25519 import Ed25519PrivateKey
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.backends import default_backend

from onionbalance.common import log
from onionbalance.common import intro_point_set

from onionbalance.hs_v3 import params

logger = log.get_logger()
backend = default_backend()


class IntroductionPointSetV3(intro_point_set.IntroductionPointSet):
    """
    This class represents a set of introduction points (which are actually
    stem.descriptor.hidden_service.IntroductionPointV3 objects)

    It gives us a nice way to compare sets of introduction poinst between them,
    to see if they are different.

    It also preserves all the functionality of
    onionbalance.common.intro_point_set.IntroductionPointSet which allows you to
    sample introduction points out of the set.
    """

    def __init__(self, introduction_points):
        """
        'introduction_points' is a list of lists where each internal list contains
        the introduction points of an instance
        """
        for instance_ips in introduction_points:
            for ip in instance_ips:
                if ip.legacy_key_raw:
                    logger.info("Ignoring introduction point with legacy key.")
                    instance_ips.remove(ip)

        super().__init__(introduction_points)

    def get_intro_points_flat(self):
        """
        Flatten the .intro_points list of list into a single list and return it
        """
        return list(itertools.chain(*self.intro_points))

    def __eq__(self, other):
        """
        Compares two IntroductionPointSetV3 objects and returns True
        if they have the same introduction points in them.
        """
        # XXX we are currently using onion_key_raw as the identifier for the
        # intro point. is there a better thing to use?
        intro_set_1 = set(ip.onion_key_raw for ip in other.get_intro_points_flat())
        intro_set_2 = set(ip.onion_key_raw for ip in self.get_intro_points_flat())

        # TODO: unittests
        return intro_set_1 == intro_set_2


class V3Descriptor(object):
    """
    A generic v3 descriptor.

    Serves as the base class for OBDescriptor and ReceivedDescriptor which
    implement more specific functionalities.
    """

    def __init__(self, onion_address, v3_desc):
        self.onion_address = onion_address

        self.v3_desc = v3_desc

        # An IntroductionPointSetV3 object with the intros of this descriptor
        self.intro_set = IntroductionPointSetV3([self.v3_desc._inner_layer.introduction_points])

    def get_intro_points(self):
        """
        Get the raw intro points for this descriptor.
        """
        return self.intro_set.get_intro_points_flat()

    def get_blinded_key(self):
        """
        Extract and return the blinded key from the descriptor
        """

        # The descriptor signing cert, signs the descriptor signing key using
        # the blinded key. So the signing key should be the one we want here.
        return self.v3_desc.signing_cert.signing_key()

    def get_size(self):
        """
        Return size of v3 descriptor in bytes
        """
        return len(str(self.v3_desc))


class OBDescriptor(V3Descriptor):
    """
    A v3 descriptor created by Onionbalance and meant to be published to the
    network.

    This class supports generating descriptors.

    Can raise BadDescriptor if we can't or should not generate a valid descriptor.
    """

    def __init__(self, onion_address, identity_priv_key,
                 blinding_param, intro_points, is_first_desc):
        # Timestamp of the last attempt to assemble this descriptor
        self.last_publish_attempt_ts = None
        # Timestamp we last uploaded this descriptor
        self.last_upload_ts = None
        # Set of responsible HSDirs for last time we uploaded this descriptor
        self.responsible_hsdirs = None

        # Start generating descriptor
        desc_signing_key = Ed25519PrivateKey.generate()

        # Get the intro points for this descriptor and recertify them!
        recertified_intro_points = []
        for ip in intro_points:
            recertified_intro_points.append(self._recertify_intro_point(ip, desc_signing_key))

        rev_counter = self._get_revision_counter(identity_priv_key, is_first_desc)

        v3_desc_inner_layer = InnerLayer.create(introduction_points=recertified_intro_points)
        v3_desc = HiddenServiceDescriptorV3.create(
            blinding_nonce=blinding_param,
            identity_key=identity_priv_key,
            signing_key=desc_signing_key,
            inner_layer=v3_desc_inner_layer,
            revision_counter=int(rev_counter),
        )

        # TODO stem should probably initialize it itself so that it has balance
        # between descriptor creation (where this is not inted) and descriptor
        # parsing (where this is inited)
        v3_desc._inner_layer = v3_desc_inner_layer

        # Check max size is within range
        if len(str(v3_desc)) > params.MAX_DESCRIPTOR_SIZE:
            logger.error("Created descriptor is too big (%d intros). Consider "
                         "relaxing number of instances or intro points per instance "
                         "(see N_INTROS_PER_INSTANCE)")
            raise BadDescriptor

        super().__init__(onion_address, v3_desc)

    def set_last_publish_attempt_ts(self, last_publish_attempt_ts):
        self.last_publish_attempt_ts = last_publish_attempt_ts

    def set_last_upload_ts(self, last_upload_ts):
        self.last_upload_ts = last_upload_ts

    def set_responsible_hsdirs(self, responsible_hsdirs):
        self.responsible_hsdirs = responsible_hsdirs

    def _recertify_intro_point(self, intro_point, descriptor_signing_key):
        """
        Given an IntroductionPointV3 object, re-certify it using the
        'descriptor_signing_key' for this new descriptor.

        Return the recertified intro point.
        """
        original_auth_key_cert = intro_point.auth_key_cert
        original_enc_key_cert = intro_point.enc_key_cert

        # We have already removed all the intros with legacy keys. Make sure that
        # no legacy intros sneaks up on us, becausey they would result in
        # unparseable descriptors if we don't recertify them (and we won't).
        assert(not intro_point.legacy_key_cert)

        # Get all the certs we need to recertify
        # [we need to use the _replace method of namedtuples because there is no
        # setter for those attributes due to the way stem sets those fields. If we
        # attempt to normally replace the attributes we get the following
        # exception: AttributeError: can't set attribute]
        recertified_intro_point = intro_point._replace(auth_key_cert=self._recertify_ed_certificate(original_auth_key_cert,
                                                                                                    descriptor_signing_key),
                                                       enc_key_cert=self._recertify_ed_certificate(original_enc_key_cert,
                                                                                                   descriptor_signing_key))

        return recertified_intro_point

    def _recertify_ed_certificate(self, ed_cert, descriptor_signing_key):
        """
        Recertify an HSv3 intro point certificate using the new descriptor signing
        key so that it can be accepted as part of a new descriptor.

        "Recertifying" means taking the certified key and signing it with a new
        key.

        Return the new certificate.
        """
        # pylint: disable=no-member
        extensions = [Ed25519Extension(ExtensionType.HAS_SIGNING_KEY, None, stem.util._pubkey_bytes(descriptor_signing_key))]
        new_cert = Ed25519CertificateV1(cert_type=ed_cert.type,
                                        expiration=ed_cert.expiration,
                                        key_type=ed_cert.key_type,
                                        key=ed_cert.key,
                                        extensions=extensions,
                                        signing_key=descriptor_signing_key)

        return new_cert

    def _get_revision_counter(self, identity_priv_key, is_first_desc):
        """
        Get the revision counter using the order-preserving-encryption scheme from
        rend-spec-v3.txt section F.2.
        """
        from onionbalance.hs_v3.onionbalance import my_onionbalance
        now = int(stem.util.datetime_to_unix(datetime.datetime.utcnow()))

        # TODO: Mention that this is done with the private key instead of the blinded priv key
        # this means that this won't cooperate with normal tor
        privkey_bytes = identity_priv_key.private_bytes(encoding=serialization.Encoding.Raw,
                                                        format=serialization.PrivateFormat.Raw,
                                                        encryption_algorithm=serialization.NoEncryption())
        cipher_key = hashlib.sha3_256(b"rev-counter-generation" + privkey_bytes).digest()

        if is_first_desc:
            srv_start = my_onionbalance.consensus.get_start_time_of_previous_srv_run()
        else:
            srv_start = my_onionbalance.consensus.get_start_time_of_current_srv_run()
        srv_start = int(srv_start)

        seconds_since_srv_start = now - srv_start
        # This must be strictly positive
        seconds_since_srv_start += 1

        ope_result = sum(w for w, _ in zip(self._get_ope_scheme_words(cipher_key),
                                           range(seconds_since_srv_start)))

        logger.debug("Rev counter for %s descriptor (SRV secs %s, OPE %s)",
                     "first" if is_first_desc else "second",
                     seconds_since_srv_start, ope_result)

        return ope_result

    def _get_ope_scheme_words(self, cipher_key):
        IV = b'\x00' * 16

        cipher = Cipher(algorithms.AES(cipher_key), modes.CTR(IV), backend=backend)
        e = cipher.encryptor()
        while True:
            v = e.update(b'\x00\x00')
            yield v[0] + 256 * v[1] + 1


class ReceivedDescriptor(V3Descriptor):
    """
    An instance v3 descriptor received from the network.

    This class supports parsing descriptors.
    """

    def __init__(self, desc_text, onion_address):
        """
        Parse a descriptor in 'desc_text' and return an ReceivedDescriptor object.

        Raises BadDescriptor if the descriptor cannot be used.
        """
        try:
            v3_desc = HiddenServiceDescriptorV3.from_str(desc_text)
            v3_desc.decrypt(onion_address)
        except ValueError as err:
            logger.warning("Descriptor is corrupted (%s).", err)
            raise BadDescriptor

        self.received_ts = datetime.datetime.utcnow()

        logger.debug("Successfuly decrypted descriptor for %s!", onion_address)

        super().__init__(onion_address, v3_desc)

    def is_old(self):
        """
        Return True if this received descriptor is old and we should consider the
        instance as offline.
        """
        from onionbalance.hs_v3.onionbalance import my_onionbalance

        received_age = datetime.datetime.utcnow() - self.received_ts
        received_age = received_age.total_seconds()

        if my_onionbalance.is_testnet:
            too_old_threshold = params.INSTANCE_DESCRIPTOR_TOO_OLD_TESTNET
        else:
            too_old_threshold = params.INSTANCE_DESCRIPTOR_TOO_OLD

        if received_age > too_old_threshold:
            return True

        return False


class BadDescriptor(Exception):
    pass



# -*- coding: utf-8 -*-
import datetime
import string

import pytest
import Cryptodome.PublicKey.RSA
import stem.descriptor
import stem.descriptor.hidden_service_descriptor
import hashlib

from binascii import unhexlify

from onionbalance.common import intro_point_set
from onionbalance.hs_v2 import descriptor

PEM_PRIVATE_KEY = u'\n'.join([
    '-----BEGIN RSA PRIVATE KEY-----',
    'MIICWwIBAAKBgQDXzP6HGtjPSy7uF9OlY7ZmefTVKcFLsq0mSEzQrW5wSiNuYc+d',
    'oSV2OWxPg+1fVe19ES43AUkq/bS/gjAMLOunP6u9FbPDojyh1Vs/6TVqftS3sPkl',
    'Q0ItrrZwAwhtHC0WaEyrwYJNOSCBq3wpupdQhpRyWJFqMwm9+iBCG1QcJQIDAQAB',
    'AoGAegc2Sqm4vgdyozof+R8Ybnw6ISu6XRbNaJ9rqHjZwW9695khsK4GJAM2pwQf',
    '/0/0ukszyfDVMhVC1yREDS59lgzNecItd6nQZWbwr9TFxIoa9ouTqk8PcAoNixTb',
    'wafjPcMmWGakizXeAHiOfazPBH4x2keDQCulxfYxXZxTpyECQQDqZu61kd1S3U7T',
    'BT2NQBd3tHX0Hvonx+IkOKXwpHFY0Mo4d32Bi+MxRuEnd3tO44AaMvlkl13QMTF2',
    'kHFSC70dAkEA669LZavGjW67+rO+f+xyDVby9pD5GJQBb78xRCf93Zcu2KW4NSp3',
    'XC4p4eWfLgff1VuXL7g0VdFm4wUUHqYUqQJAZLmqpjdyBeO3tZIw6vu5meTgMvEE',
    'ygdos+vr0sa3NlUyMKWYNwznqgstQYpkYHf+WkPBS2qIE6iv+qUDLSCCOQJAESSk',
    'CFYxUBJQ7BBs9+Mb/Kppa9Ppuobxf85ZaAq8pYScrLeJKZzYJ8VX2I2aQX/jISLT',
    'YW41qFRd9n9lEkGkWQJAcxPmNI+2r5zJG+K148LLmWCIDTVZ4nxOcxffHka/3tCJ',
    'lDGUw4p2wU6pVRDpNfKrF5Nc9ZKO8NAtC17ZvDyVkQ==',
    '-----END RSA PRIVATE KEY-----',
])

INTRODUCTION_POINT_PART = u'\n'.join([
    '-----BEGIN MESSAGE-----',
    'AgEdbps604RR6lqeyoZBzOb6+HvlL2cDt63w8vBtyRaLirq5ZD5GDnr+R0ePj71C',
    'nC7qmRWuwBmzSdSd0lOTaSApBvIifbJksHUeT/rq03dpnnRHdHSVqSvig6bukcWJ',
    'LgJmrRd3ES13LXVHenD3C6AZMHuL9TG+MjLO2PIHu0mFO18aAHVnWY32Dmt144IY',
    'c2eTVZbsKobjjwCYvDf0PBZI+B6H0PZWkDX/ykYjArpLDwydeZyp+Zwj4+k0+nRr',
    'RPlzbHYoBY9pFYDUXDXWdL+vTsgFTG0EngLGlgUWSY5U1T1Db5HfOqc7hbqklgs/',
    'ULG8NUY1k41Wb+dleJI28/+ZOM9zOpHcegNx4Cn8UGbw/Yv3Tj+yki+TMeOtJyhK',
    'PQP8NWq8zThiVhBrfpmVjMYkNeVNyVNoxRwS6rxCQjoLWSJit2Mpf57zY1AOvT1S',
    'EqqFbsX+slD2Uk67imALh4pMtjX29VLIujpum3drLhoTHDszBRhIH61A2eAZqdJy',
    '7JkJd1x/8x7U0l8xNWhnj/bhUHdt3OrCvlN+n8x6BwmMNoLF8JIsskTuGHOaAKSQ',
    'WK3z0rHjgIrEjkQeuQtfmptiIgRB9LnNr+YahRnRR6XIOJGaIoVLVM2Uo2RG4MS1',
    '2KC3DRJ87WdMv2yNWha3w+lWt/mOALahYrvuNMU8wEuNXSi5yCo1OKirv+d5viGe',
    'hAgVZjRymBQF+vd30zMdOG9qXNoQFUN49JfS8z5FjWmdHRt2MHlqD2isxoeabERY',
    'T4Q50fFH8XHkRRomKBEbCwy/4t2DiqcTOSLGOSbTtf7qlUACp2bRth/g0ySAW8X/',
    'CaWVm53z1vdgF2+t6j1CnuIqf0dUygZ07HEAHgu3rMW0YTk04QkvR3jiKAKijvGH',
    '3YcMJz1aJ7psWSsgiwn8a8Cs4fAcLNJcdTrnyxhQI4PMST/QLfp8nPYrhKEeifTc',
    'vYkC4CtGuEFkWyRifIGbeD7FcjkL1zqVNu31vgo3EIVbHzylERgpgTIYBRv7aV7W',
    'X7XAbrrgXL0zgpI0orOyPkr2KRs6CcoEqcc2MLyB6gJ5fYAm69Ige+6gWtRT6qvZ',
    'tJXagfKZivLj73dRD6sUqTCX4tmgo7Q8WFSeNscDAVm/p4dVsw6SOoFcRgaH20yX',
    'MBa3oLNTUNAaGbScUPx2Ja3MQS0UITwk0TFTF7hL++NhTvTp6IdgQW4DG+/bVJ3M',
    'BRR+hsvSz5BSQQj2FUIAsJ+WoVK9ImbgsBbYxSH60jCvxTIdeh2IeUzS2T1bU9AU',
    'jOLzcJZmNh95Nj2Qdrc8/0gin9KpgPmuPQ6CyH3TPFy88lf19v9jHUMO4SKEr7am',
    'DAjbX3D7APKgHyZ61CkuoB3gylIRb8rRJD2ote38M6A1+04yJL/jG+PCL1UnMWdL',
    'yJ4f4LzI9c4ksnGyl9neq0IHnA0Nlky6dmgmE+vLi6OCbEEs2v132wc5PIxRY+TW',
    '8JWu+3wUA4tj5uQvQRqU9/lmoHG/Jxubx/HwdD9Ri17G+qX8re5sySmmq7rcZEGJ',
    'LVrlFuvA0NdoTM4AZY23iR6trJ/Ba2Q4pQk4SfOEMSoZJmf0UbxIP0Ez6Fb+Dxzk',
    'WKXfI+D0ScuVjzV0bs8iXTrCcynztRKndNbtpd39hGAR0rNqvnHyQGYV75bWm5dS',
    '0S0PQ6DOzicLxjNXZFicQvwfieg9VyJikWLFLu4zAbzHnuoRk6b2KbSU4UCG/BCz',
    'mHqz4y6GfsncsNkmFmsD5Gn9UrloWcEWgIDL05yIikL+L9DPLnNlSYtehDfxlhvh',
    'xHzY/Rad4Nzxe62yXhSxhROLTXIolllyOFJgqZ4hBlXybBqJH7sZUll6PUpDwZdu',
    'BK14pzMIpfxq2eYp8jI7fh4lU9YrkuSUM0Ewa7HfrltAgxMhHyaFjfINt61P9OlO',
    's3nuBY17+KokaSWjACkCimVLH13H5DRhfX8OBRT4LeRMUspX3cyKbccwpOmoBf4y',
    'WPM9QXw7nQy2hwnuX6NiK5QfeCGfY64M06J2tBGcCDmjPSIcJgMcyY7jfH9yPlDt',
    'SKyyXpZnFOJplS2v28A/1csPSGy9kk/uGN0hfFULH4VvyAgNDYzmeOd8FvrbfHH2',
    '8BUTI/Tq2pckxwCYBWHcjSdXRAj5moCNSxCUMtK3kWFdxLFYzoiKuiZwq171qb5L',
    'yCHMwNDIWEMeC75XSMswHaBsK6ON0UUg5oedQkOK+II9L/DVyTs3UYJOsWDfM67E',
    '312O9/bmsoHvr+rofF7HEc74dtUAcaDGJNyNiB+O4UmWbtEpCfuLmq2vaZa9J7Y0',
    'hXlD2pcibC9CWpKR58cRL+dyYHZGJ4VKg6OHlJlF+JBPeLzObNDz/zQuEt9aL9Ae',
    'QByamqGDGcaVMVZ/A80fRoUUgHbh3bLoAmxLCvMbJ0YMtRujdtGm8ZD0WvLXQA/U',
    'dNmQ6tsP6pyVorWVa/Ma5CR7Em5q7M6639T8WPcu7ETTO19MnWud2lPJ5A==',
    '-----END MESSAGE-----',
])

UNSIGNED_DESCRIPTOR = u'\n'.join([
    'rendezvous-service-descriptor 6wgohrr64y2od75psnrfdkbc74ddqx2v',
    'version 2',
    'permanent-key',
    '-----BEGIN RSA PUBLIC KEY-----',
    'MIGJAoGBANfM/oca2M9LLu4X06VjtmZ59NUpwUuyrSZITNCtbnBKI25hz52hJXY5',
    'bE+D7V9V7X0RLjcBSSr9tL+CMAws66c/q70Vs8OiPKHVWz/pNWp+1Lew+SVDQi2u',
    'tnADCG0cLRZoTKvBgk05IIGrfCm6l1CGlHJYkWozCb36IEIbVBwlAgMBAAE=',
    '-----END RSA PUBLIC KEY-----',
    'secret-id-part udmoj3e2ykfp73kpvauoq4t4p7kkwsjq',
    'publication-time 2015-06-25 11:00:00',
    'protocol-versions 2,3',
    'introduction-points',
    '-----BEGIN MESSAGE-----',
    'AgEdbps604RR6lqeyoZBzOb6+HvlL2cDt63w8vBtyRaLirq5ZD5GDnr+R0ePj71C',
    'nC7qmRWuwBmzSdSd0lOTaSApBvIifbJksHUeT/rq03dpnnRHdHSVqSvig6bukcWJ',
    'LgJmrRd3ES13LXVHenD3C6AZMHuL9TG+MjLO2PIHu0mFO18aAHVnWY32Dmt144IY',
    'c2eTVZbsKobjjwCYvDf0PBZI+B6H0PZWkDX/ykYjArpLDwydeZyp+Zwj4+k0+nRr',
    'RPlzbHYoBY9pFYDUXDXWdL+vTsgFTG0EngLGlgUWSY5U1T1Db5HfOqc7hbqklgs/',
    'ULG8NUY1k41Wb+dleJI28/+ZOM9zOpHcegNx4Cn8UGbw/Yv3Tj+yki+TMeOtJyhK',
    'PQP8NWq8zThiVhBrfpmVjMYkNeVNyVNoxRwS6rxCQjoLWSJit2Mpf57zY1AOvT1S',
    'EqqFbsX+slD2Uk67imALh4pMtjX29VLIujpum3drLhoTHDszBRhIH61A2eAZqdJy',
    '7JkJd1x/8x7U0l8xNWhnj/bhUHdt3OrCvlN+n8x6BwmMNoLF8JIsskTuGHOaAKSQ',
    'WK3z0rHjgIrEjkQeuQtfmptiIgRB9LnNr+YahRnRR6XIOJGaIoVLVM2Uo2RG4MS1',
    '2KC3DRJ87WdMv2yNWha3w+lWt/mOALahYrvuNMU8wEuNXSi5yCo1OKirv+d5viGe',
    'hAgVZjRymBQF+vd30zMdOG9qXNoQFUN49JfS8z5FjWmdHRt2MHlqD2isxoeabERY',
    'T4Q50fFH8XHkRRomKBEbCwy/4t2DiqcTOSLGOSbTtf7qlUACp2bRth/g0ySAW8X/',
    'CaWVm53z1vdgF2+t6j1CnuIqf0dUygZ07HEAHgu3rMW0YTk04QkvR3jiKAKijvGH',
    '3YcMJz1aJ7psWSsgiwn8a8Cs4fAcLNJcdTrnyxhQI4PMST/QLfp8nPYrhKEeifTc',
    'vYkC4CtGuEFkWyRifIGbeD7FcjkL1zqVNu31vgo3EIVbHzylERgpgTIYBRv7aV7W',
    'X7XAbrrgXL0zgpI0orOyPkr2KRs6CcoEqcc2MLyB6gJ5fYAm69Ige+6gWtRT6qvZ',
    'tJXagfKZivLj73dRD6sUqTCX4tmgo7Q8WFSeNscDAVm/p4dVsw6SOoFcRgaH20yX',
    'MBa3oLNTUNAaGbScUPx2Ja3MQS0UITwk0TFTF7hL++NhTvTp6IdgQW4DG+/bVJ3M',
    'BRR+hsvSz5BSQQj2FUIAsJ+WoVK9ImbgsBbYxSH60jCvxTIdeh2IeUzS2T1bU9AU',
    'jOLzcJZmNh95Nj2Qdrc8/0gin9KpgPmuPQ6CyH3TPFy88lf19v9jHUMO4SKEr7am',
    'DAjbX3D7APKgHyZ61CkuoB3gylIRb8rRJD2ote38M6A1+04yJL/jG+PCL1UnMWdL',
    'yJ4f4LzI9c4ksnGyl9neq0IHnA0Nlky6dmgmE+vLi6OCbEEs2v132wc5PIxRY+TW',
    '8JWu+3wUA4tj5uQvQRqU9/lmoHG/Jxubx/HwdD9Ri17G+qX8re5sySmmq7rcZEGJ',
    'LVrlFuvA0NdoTM4AZY23iR6trJ/Ba2Q4pQk4SfOEMSoZJmf0UbxIP0Ez6Fb+Dxzk',
    'WKXfI+D0ScuVjzV0bs8iXTrCcynztRKndNbtpd39hGAR0rNqvnHyQGYV75bWm5dS',
    '0S0PQ6DOzicLxjNXZFicQvwfieg9VyJikWLFLu4zAbzHnuoRk6b2KbSU4UCG/BCz',
    'mHqz4y6GfsncsNkmFmsD5Gn9UrloWcEWgIDL05yIikL+L9DPLnNlSYtehDfxlhvh',
    'xHzY/Rad4Nzxe62yXhSxhROLTXIolllyOFJgqZ4hBlXybBqJH7sZUll6PUpDwZdu',
    'BK14pzMIpfxq2eYp8jI7fh4lU9YrkuSUM0Ewa7HfrltAgxMhHyaFjfINt61P9OlO',
    's3nuBY17+KokaSWjACkCimVLH13H5DRhfX8OBRT4LeRMUspX3cyKbccwpOmoBf4y',
    'WPM9QXw7nQy2hwnuX6NiK5QfeCGfY64M06J2tBGcCDmjPSIcJgMcyY7jfH9yPlDt',
    'SKyyXpZnFOJplS2v28A/1csPSGy9kk/uGN0hfFULH4VvyAgNDYzmeOd8FvrbfHH2',
    '8BUTI/Tq2pckxwCYBWHcjSdXRAj5moCNSxCUMtK3kWFdxLFYzoiKuiZwq171qb5L',
    'yCHMwNDIWEMeC75XSMswHaBsK6ON0UUg5oedQkOK+II9L/DVyTs3UYJOsWDfM67E',
    '312O9/bmsoHvr+rofF7HEc74dtUAcaDGJNyNiB+O4UmWbtEpCfuLmq2vaZa9J7Y0',
    'hXlD2pcibC9CWpKR58cRL+dyYHZGJ4VKg6OHlJlF+JBPeLzObNDz/zQuEt9aL9Ae',
    'QByamqGDGcaVMVZ/A80fRoUUgHbh3bLoAmxLCvMbJ0YMtRujdtGm8ZD0WvLXQA/U',
    'dNmQ6tsP6pyVorWVa/Ma5CR7Em5q7M6639T8WPcu7ETTO19MnWud2lPJ5A==',
    '-----END MESSAGE-----',
    'signature',
    '-----BEGIN SIGNATURE-----',
    'VX4GC6s6zmY84mKsh+YdAqyZqDevJwGYr9yJntBNms4XRQHlgiW/JCspJzCqvrQG',
    'N4Fh8XNTodQFnxz/kz8K3SBFlLnJHzKxSBTSZTLd8hRp84F/XxDcPaIPda8UJZuF',
    'pOT8V0hfhgo8WxLpOyUzxrYugPB2GRkWYLhHaKhxkJY=',
    '-----END SIGNATURE-----',
])

SIGNED_DESCRIPTOR = u'\n'.join([
    'rendezvous-service-descriptor 6wgohrr64y2od75psnrfdkbc74ddqx2v',
    'version 2',
    'permanent-key',
    '-----BEGIN RSA PUBLIC KEY-----',
    'MIGJAoGBANfM/oca2M9LLu4X06VjtmZ59NUpwUuyrSZITNCtbnBKI25hz52hJXY5',
    'bE+D7V9V7X0RLjcBSSr9tL+CMAws66c/q70Vs8OiPKHVWz/pNWp+1Lew+SVDQi2u',
    'tnADCG0cLRZoTKvBgk05IIGrfCm6l1CGlHJYkWozCb36IEIbVBwlAgMBAAE=',
    '-----END RSA PUBLIC KEY-----',
    'secret-id-part udmoj3e2ykfp73kpvauoq4t4p7kkwsjq',
    'publication-time 2015-06-25 11:00:00',
    'protocol-versions 2,3',
    'introduction-points',
    '-----BEGIN MESSAGE-----',
    'AgEdbps604RR6lqeyoZBzOb6+HvlL2cDt63w8vBtyRaLirq5ZD5GDnr+R0ePj71C',
    'nC7qmRWuwBmzSdSd0lOTaSApBvIifbJksHUeT/rq03dpnnRHdHSVqSvig6bukcWJ',
    'LgJmrRd3ES13LXVHenD3C6AZMHuL9TG+MjLO2PIHu0mFO18aAHVnWY32Dmt144IY',
    'c2eTVZbsKobjjwCYvDf0PBZI+B6H0PZWkDX/ykYjArpLDwydeZyp+Zwj4+k0+nRr',
    'RPlzbHYoBY9pFYDUXDXWdL+vTsgFTG0EngLGlgUWSY5U1T1Db5HfOqc7hbqklgs/',
    'ULG8NUY1k41Wb+dleJI28/+ZOM9zOpHcegNx4Cn8UGbw/Yv3Tj+yki+TMeOtJyhK',
    'PQP8NWq8zThiVhBrfpmVjMYkNeVNyVNoxRwS6rxCQjoLWSJit2Mpf57zY1AOvT1S',
    'EqqFbsX+slD2Uk67imALh4pMtjX29VLIujpum3drLhoTHDszBRhIH61A2eAZqdJy',
    '7JkJd1x/8x7U0l8xNWhnj/bhUHdt3OrCvlN+n8x6BwmMNoLF8JIsskTuGHOaAKSQ',
    'WK3z0rHjgIrEjkQeuQtfmptiIgRB9LnNr+YahRnRR6XIOJGaIoVLVM2Uo2RG4MS1',
    '2KC3DRJ87WdMv2yNWha3w+lWt/mOALahYrvuNMU8wEuNXSi5yCo1OKirv+d5viGe',
    'hAgVZjRymBQF+vd30zMdOG9qXNoQFUN49JfS8z5FjWmdHRt2MHlqD2isxoeabERY',
    'T4Q50fFH8XHkRRomKBEbCwy/4t2DiqcTOSLGOSbTtf7qlUACp2bRth/g0ySAW8X/',
    'CaWVm53z1vdgF2+t6j1CnuIqf0dUygZ07HEAHgu3rMW0YTk04QkvR3jiKAKijvGH',
    '3YcMJz1aJ7psWSsgiwn8a8Cs4fAcLNJcdTrnyxhQI4PMST/QLfp8nPYrhKEeifTc',
    'vYkC4CtGuEFkWyRifIGbeD7FcjkL1zqVNu31vgo3EIVbHzylERgpgTIYBRv7aV7W',
    'X7XAbrrgXL0zgpI0orOyPkr2KRs6CcoEqcc2MLyB6gJ5fYAm69Ige+6gWtRT6qvZ',
    'tJXagfKZivLj73dRD6sUqTCX4tmgo7Q8WFSeNscDAVm/p4dVsw6SOoFcRgaH20yX',
    'MBa3oLNTUNAaGbScUPx2Ja3MQS0UITwk0TFTF7hL++NhTvTp6IdgQW4DG+/bVJ3M',
    'BRR+hsvSz5BSQQj2FUIAsJ+WoVK9ImbgsBbYxSH60jCvxTIdeh2IeUzS2T1bU9AU',
    'jOLzcJZmNh95Nj2Qdrc8/0gin9KpgPmuPQ6CyH3TPFy88lf19v9jHUMO4SKEr7am',
    'DAjbX3D7APKgHyZ61CkuoB3gylIRb8rRJD2ote38M6A1+04yJL/jG+PCL1UnMWdL',
    'yJ4f4LzI9c4ksnGyl9neq0IHnA0Nlky6dmgmE+vLi6OCbEEs2v132wc5PIxRY+TW',
    '8JWu+3wUA4tj5uQvQRqU9/lmoHG/Jxubx/HwdD9Ri17G+qX8re5sySmmq7rcZEGJ',
    'LVrlFuvA0NdoTM4AZY23iR6trJ/Ba2Q4pQk4SfOEMSoZJmf0UbxIP0Ez6Fb+Dxzk',
    'WKXfI+D0ScuVjzV0bs8iXTrCcynztRKndNbtpd39hGAR0rNqvnHyQGYV75bWm5dS',
    '0S0PQ6DOzicLxjNXZFicQvwfieg9VyJikWLFLu4zAbzHnuoRk6b2KbSU4UCG/BCz',
    'mHqz4y6GfsncsNkmFmsD5Gn9UrloWcEWgIDL05yIikL+L9DPLnNlSYtehDfxlhvh',
    'xHzY/Rad4Nzxe62yXhSxhROLTXIolllyOFJgqZ4hBlXybBqJH7sZUll6PUpDwZdu',
    'BK14pzMIpfxq2eYp8jI7fh4lU9YrkuSUM0Ewa7HfrltAgxMhHyaFjfINt61P9OlO',
    's3nuBY17+KokaSWjACkCimVLH13H5DRhfX8OBRT4LeRMUspX3cyKbccwpOmoBf4y',
    'WPM9QXw7nQy2hwnuX6NiK5QfeCGfY64M06J2tBGcCDmjPSIcJgMcyY7jfH9yPlDt',
    'SKyyXpZnFOJplS2v28A/1csPSGy9kk/uGN0hfFULH4VvyAgNDYzmeOd8FvrbfHH2',
    '8BUTI/Tq2pckxwCYBWHcjSdXRAj5moCNSxCUMtK3kWFdxLFYzoiKuiZwq171qb5L',
    'yCHMwNDIWEMeC75XSMswHaBsK6ON0UUg5oedQkOK+II9L/DVyTs3UYJOsWDfM67E',
    '312O9/bmsoHvr+rofF7HEc74dtUAcaDGJNyNiB+O4UmWbtEpCfuLmq2vaZa9J7Y0',
    'hXlD2pcibC9CWpKR58cRL+dyYHZGJ4VKg6OHlJlF+JBPeLzObNDz/zQuEt9aL9Ae',
    'QByamqGDGcaVMVZ/A80fRoUUgHbh3bLoAmxLCvMbJ0YMtRujdtGm8ZD0WvLXQA/U',
    'dNmQ6tsP6pyVorWVa/Ma5CR7Em5q7M6639T8WPcu7ETTO19MnWud2lPJ5A==',
    '-----END MESSAGE-----',
    'signature',
    '-----BEGIN SIGNATURE-----',
    'VX4GC6s6zmY84mKsh+YdAqyZqDevJwGYr9yJntBNms4XRQHlgiW/JCspJzCqvrQG',
    'N4Fh8XNTodQFnxz/kz8K3SBFlLnJHzKxSBTSZTLd8hRp84F/XxDcPaIPda8UJZuF',
    'pOT8V0hfhgo8WxLpOyUzxrYugPB2GRkWYLhHaKhxkJY=',
    '-----END SIGNATURE-----',
])

PRIVATE_KEY = Cryptodome.PublicKey.RSA.importKey(PEM_PRIVATE_KEY)
UNIX_TIMESTAMP = 1435233021


"""
TODO: Reenable test that fails with Pytest3
@pytest.mark.parametrize('intro_point_distribution, selected_ip_count', [
    ([3], 3),
    ([3, 3], 6),
    ([0], 0),
    ([10, 10], 10),
    ([3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3], 10),
    ([10, 10, 10, 10, 10, 10], 10),
    pytest.mark.xfail(([0, 3, 3], 10)),
    pytest.mark.xfail(([6, 3, 3], 12)),
])
def test_introduction_point_selection(intro_point_distribution,
                                      selected_ip_count):
    # Basic test case to check that the correct number of IPs are selected.
    # Create Mock list of instances (index by letter) and their respective
    # introduction points.
    available_intro_points = [[index] * count for index, count
                              in zip(string.ascii_lowercase,
                                     intro_point_distribution)]

    intro_set = intro_point_set.IntroductionPointSet(available_intro_points)

    # Check that we can fetch the same number for each descriptor
    for i in range(0, 2):
        # Max 10 introduction points per descriptor
        choosen_intro_points = intro_set.choose(10)
        assert len(choosen_intro_points) == selected_ip_count
"""

def test_generate_service_descriptor(monkeypatch, mocker):
    """
    Test creation of a fully signed hidden service descriptor
    """
    # Mock the datetime function to return a constant timestamp
    class frozen_datetime(datetime.datetime):
        @classmethod
        def utcnow(cls):
            return datetime.datetime.utcfromtimestamp(UNIX_TIMESTAMP)
    monkeypatch.setattr(datetime, 'datetime', frozen_datetime)

    # Patch make_introduction_points_part to return the test introduction
    # point section
    mocker.patch('onionbalance.hs_v2.descriptor.make_introduction_points_part',
                 lambda *_: INTRODUCTION_POINT_PART)

    # Test basic descriptor generation.
    signed_descriptor = descriptor.generate_service_descriptor(
        PRIVATE_KEY,
        introduction_point_list=['mocked-ip-list'],
    ).encode('utf-8')
    stem.descriptor.hidden_service_descriptor.\
        HiddenServiceDescriptor(signed_descriptor, validate=True)
    assert (hashlib.sha1(signed_descriptor).hexdigest() ==
            'df4f4a7a15492205f073c32cbcfc4eb9511e4ad8')

    # Test descriptor generation with specified timestamp
    signed_descriptor = descriptor.generate_service_descriptor(
        PRIVATE_KEY,
        introduction_point_list=['mocked-ip-list'],
        timestamp=datetime.datetime.utcfromtimestamp(UNIX_TIMESTAMP),
    ).encode('utf-8')
    stem.descriptor.hidden_service_descriptor.\
        HiddenServiceDescriptor(signed_descriptor, validate=True)
    assert (hashlib.sha1(signed_descriptor).hexdigest() ==
            'df4f4a7a15492205f073c32cbcfc4eb9511e4ad8')

    # Test descriptor for deviation and replica 1
    signed_descriptor = descriptor.generate_service_descriptor(
        PRIVATE_KEY,
        introduction_point_list=['mocked-ip-list'],
        replica=1,
        deviation=24*60*60,
    ).encode('utf-8')
    stem.descriptor.hidden_service_descriptor.\
        HiddenServiceDescriptor(signed_descriptor, validate=True)
    assert (hashlib.sha1(signed_descriptor).hexdigest() ==
            'd828140cdccb1165dbc5a4b39622fcb45e6438fb')


def test_generate_service_descriptor_no_intros():
    with pytest.raises(ValueError):
        descriptor.generate_service_descriptor(
            PRIVATE_KEY,
            introduction_point_list=[],
        )


def test_make_public_key_block():
    """
    Test generation of ASN.1 representation of public key
    """
    public_key_block = descriptor.make_public_key_block(PRIVATE_KEY)
    assert (hashlib.sha1(public_key_block.encode('utf-8')).hexdigest() ==
            '2cf75da5e1a198ca7cb3db7b0baa6708feaf26e8')

def test_sign_descriptor():
    """
    Test signing a descriptor
    """

    # Test signing an unsigned descriptor
    signed_descriptor = descriptor.sign_descriptor(
        UNSIGNED_DESCRIPTOR, PRIVATE_KEY).encode('utf-8')
    stem.descriptor.hidden_service_descriptor.\
        HiddenServiceDescriptor(signed_descriptor, validate=True)
    assert (hashlib.sha1(signed_descriptor).hexdigest() ==
            'df4f4a7a15492205f073c32cbcfc4eb9511e4ad8')

    # Test resigning a previously signed descriptor
    signed_descriptor = descriptor.sign_descriptor(
        SIGNED_DESCRIPTOR, PRIVATE_KEY).encode('utf-8')
    stem.descriptor.hidden_service_descriptor.\
        HiddenServiceDescriptor(signed_descriptor, validate=True)
    assert (hashlib.sha1(signed_descriptor).hexdigest() ==
            'df4f4a7a15492205f073c32cbcfc4eb9511e4ad8')


def test_descriptor_received_invalid_descriptor(mocker):
    """
    Test invalid descriptor content received from the HSDir
    """
    mocker.patch("onionbalance.hs_v2.descriptor.logger.exception",
                 side_effect=ValueError('InvalidDescriptorException'))

    # Check that the invalid descriptor error is logged.
    with pytest.raises(ValueError):
        descriptor.descriptor_received(u'not-a-valid-descriptor-input')
    assert descriptor.logger.exception.call_count == 1



import base64

import hashlib

from onionbalance.common import log

logger = log.get_logger()


class Node(object):
    """
    Represents a Tor node.

    A Node instance gets created for each node of a consensus. When we fetch a
    new consensus, we create new Node instances for the routers found inside.

    The 'microdescriptor' and 'routerstatus' fields of this object are
    immutable: They are set once when we receive the consensus based on the
    state of the network at that point, and they stay like that until we get a
    new consensus.
    """

    def __init__(self, microdescriptor, routerstatus):
        assert(microdescriptor and routerstatus)

        logger.debug("Initializing node with fpr %s", routerstatus.fingerprint)

        # The microdescriptor of this node
        self.microdescriptor = microdescriptor
        # The consensus routerstatus for this node
        self.routerstatus = routerstatus

    def get_hex_fingerprint(self):
        return self.routerstatus.fingerprint

    def get_hsdir_index(self, srv, period_num):
        """
        Get the HSDir index for this node:

           hsdir_index(node) = H("node-idx" | node_identity |
                                 shared_random_value |
                                 INT_8(period_num) |
                                 INT_8(period_length) )

        Raises NoHSDir or NoEd25519Identity in case of errors.
        """
        from onionbalance.hs_v3.onionbalance import my_onionbalance

        # See if this node can be an HSDir (it needs to be supported both in
        # protover and in flags)
        if 'HSDir' not in self.routerstatus.protocols or \
           2 not in self.routerstatus.protocols['HSDir'] or \
           'HSDir' not in self.routerstatus.flags:
            raise NoHSDir

        # See if ed25519 identity is supported for this node
        if 'ed25519' not in self.microdescriptor.identifiers:
            raise NoEd25519Identity

        # In stem the ed25519 identity is a base64 string and we need to add
        # the missing padding so that the python base64 module can successfuly
        # decode it.
        # TODO: Abstract this into its own function...
        ed25519_node_identity_b64 = self.microdescriptor.identifiers['ed25519']
        missing_padding = len(ed25519_node_identity_b64) % 4
        ed25519_node_identity_b64 += '=' * missing_padding
        ed25519_node_identity = base64.b64decode(ed25519_node_identity_b64)

        period_num_int_8 = period_num.to_bytes(8, 'big')
        period_length = my_onionbalance.consensus.get_time_period_length()
        period_length_int_8 = period_length.to_bytes(8, 'big')

        hash_body = b"%s%s%s%s%s" % (b"node-idx",
                                     ed25519_node_identity,
                                     srv,
                                     period_num_int_8, period_length_int_8)
        hsdir_index = hashlib.sha3_256(hash_body).digest()

        return hsdir_index


class NoEd25519Identity(Exception):
    pass


class NoHSDir(Exception):
    pass



# -*- coding: utf-8 -*-
from binascii import hexlify, unhexlify
import base64
import datetime
import io
import sys

import Cryptodome.PublicKey.RSA

import pytest
from .util import builtin

from onionbalance.hs_v2.util import *
from onionbalance.hs_v2.descriptor import pad_msg_with_tor_pkcs


PEM_PRIVATE_KEY = u'\n'.join([
    "-----BEGIN RSA PRIVATE KEY-----",
    "MIICWwIBAAKBgQDXzP6HGtjPSy7uF9OlY7ZmefTVKcFLsq0mSEzQrW5wSiNuYc+d",
    "oSV2OWxPg+1fVe19ES43AUkq/bS/gjAMLOunP6u9FbPDojyh1Vs/6TVqftS3sPkl",
    "Q0ItrrZwAwhtHC0WaEyrwYJNOSCBq3wpupdQhpRyWJFqMwm9+iBCG1QcJQIDAQAB",
    "AoGAegc2Sqm4vgdyozof+R8Ybnw6ISu6XRbNaJ9rqHjZwW9695khsK4GJAM2pwQf",
    "/0/0ukszyfDVMhVC1yREDS59lgzNecItd6nQZWbwr9TFxIoa9ouTqk8PcAoNixTb",
    "wafjPcMmWGakizXeAHiOfazPBH4x2keDQCulxfYxXZxTpyECQQDqZu61kd1S3U7T",
    "BT2NQBd3tHX0Hvonx+IkOKXwpHFY0Mo4d32Bi+MxRuEnd3tO44AaMvlkl13QMTF2",
    "kHFSC70dAkEA669LZavGjW67+rO+f+xyDVby9pD5GJQBb78xRCf93Zcu2KW4NSp3",
    "XC4p4eWfLgff1VuXL7g0VdFm4wUUHqYUqQJAZLmqpjdyBeO3tZIw6vu5meTgMvEE",
    "ygdos+vr0sa3NlUyMKWYNwznqgstQYpkYHf+WkPBS2qIE6iv+qUDLSCCOQJAESSk",
    "CFYxUBJQ7BBs9+Mb/Kppa9Ppuobxf85ZaAq8pYScrLeJKZzYJ8VX2I2aQX/jISLT",
    "YW41qFRd9n9lEkGkWQJAcxPmNI+2r5zJG+K148LLmWCIDTVZ4nxOcxffHka/3tCJ",
    "lDGUw4p2wU6pVRDpNfKrF5Nc9ZKO8NAtC17ZvDyVkQ==",
    "-----END RSA PRIVATE KEY-----",
])

PEM_INVALID_KEY = u'\n'.join([
    "-----BEGIN RSA PRIVATE KEY-----",
    "MIICWwIBAAKBgQDXzP6HGtjPSy7uF9OlY7ZmefTVKcFLsq0mSEzQrW5wSiNuYc+d",
    "oSV2OWxPg+1fVe19ES43AUkq/bS/gjAMLOunP6u9FbPDojyh1Vs/6TVqftS3sPkl",
    "Q0ItrrZwAwhtHC0WaEyrwYJNOSCBq3wpupdQhpRyWJFqMwm9+iBCG1QcJQIDAQAB",
    "AoGAegc2Sqm4vgdyozof+R8Ybnw6ISu6XRbNaJ9rqHjZwW9695khsK4GJAM2pwQf",
    "/0/0ukszyfDVMhVC1yREDS59lgzNecItd6nQZWbwr9TFxIoa9ouTqk8PcAoNixTb",
    "wafjPcMmWGakizXeAHiOfazPBH4x2keDQCulxfYxXZxTpyECQQDqZu61kd1S3U7T",
    "BT2NQBd3t          This is an invalid key             lkl13QMTF2",
    "kHFSC70dAkEA669LZavGjW67+rO+f+xyDVby9pD5GJQBb78xRCf93Zcu2KW4NSp3",
    "XC4p4eWfLgff1VuXL7g0VdFm4wUUHqYUqQJAZLmqpjdyBeO3tZIw6vu5meTgMvEE",
    "ygdos+vr0sa3NlUyMKWYNwznqgstQYpkYHf+WkPBS2qIE6iv+qUDLSCCOQJAESSk",
    "CFYxUBJQ7BBs9+Mb/Kppa9Ppuobxf85ZaAq8pYScrLeJKZzYJ8VX2I2aQX/jISLT",
    "YW41qFRd9n9lEkGkWQJAcxPmNI+2r5zJG+K148LLmWCIDTVZ4nxOcxffHka/3tCJ",
    "lDGUw4p2wU6pVRDpNfKrF5Nc9ZKO8NAtC17ZvDyVkQ==",
    "-----END RSA PRIVATE KEY-----",
])

# Private key encrypted with the password 'password'
PEM_ENCRYPTED = u'\n'.join([
    "-----BEGIN RSA PRIVATE KEY-----",
    "Proc-Type: 4,ENCRYPTED",
    "DEK-Info: DES-EDE3-CBC,7CB7069233655F1A",
    "",
    "EpKWFhHefxQLlKS1M6fPXLUVW0gcrHwYNd2q/0J4emhrHmO50KTC6/nVGTvYS1VC",
    "XQwzlla04Ed7kAuP7nkbvT+/6fS72iZmIO/kuhihjaMmRV+peznjEroErndRzWko",
    "LCpe70/yMrHhULGR1lLINe+dZddESfYRoGEM1IYhPEEchXZBdqThvaThgeyVmoAV",
    "A5qhBOP4QFPSV4J0Jd28wTy+uPmGgCjvfvXjx4JZ2LAfPnLXOoKotRqb/cOtMapp",
    "9EmsvjRZH3OLreeQm1BmVzcXGgHLIZWmybGNAW/M0seqeD+NRPXEACOBahXZsSwd",
    "krnWALTkcfLw4NXgaHKdsogDV7gWlwkXr05CrSim0+zvg+hQpVp6Phg9qrT3Jh8g",
    "988v4Fx/rlVdEpfEeXAmLUpXH3jjeyU1ZOyi8c91Vobxe1dJ9G9P8YBBqnZo1xDa",
    "q89FR852v2DKR3xv+GRpzFM43NlWLck9DcNcqIUpbrGd0qRA1k87ZwYSiUPhBvtJ",
    "dix6XfeqbqVMYiH0K4sEyuXxJ98UqFzNY3bBi9oqvoQWpo0qrRYAzHrmDg/hJbO6",
    "aw8yhe922zw8W9+IQIy2j+ZKkaHSMKqjkIwFxmig5EA+mHNDIP4HwlCxA2e2w6HG",
    "ykLE01aHMeS72qRdLVwjib4q2iTEXZVnuyFg/wVprLmLY512iWr03kbj2CVN836b",
    "vEpVIvSj5W6oNXjm+hkKA1AMcHVK96y8Ms3BtarDe4tQDh7GjipkoSXrv+2lIl0o",
    "XjumCv4Gs63Fv3kUr+jo9N3P0SGe1GggX6MOYIcZF0I=",
    "-----END RSA PRIVATE KEY-----",
])

PRIVATE_KEY = Cryptodome.PublicKey.RSA.importKey(PEM_PRIVATE_KEY)
UNIX_TIMESTAMP = 1435229421


def test_add_pkcs1_padding():
    hasher = Cryptodome.Hash.SHA1.new(b"got me thinking")
    padded_message = pad_msg_with_tor_pkcs(hasher, 128)

    """
    $ echo -n "got me thinking"| sha1sum
    65a4a767b502df42594acf2de0a0cee109ba338e  -
    """

    assert len(padded_message) == 128
    assert (padded_message == unhexlify(
        b'0001ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff'
        b'ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff'
        b'ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff'
        b'ffffffffffffffffffffff0065a4a767b502df42594acf2de0a0cee109ba338e'
    ))


def test_get_asn1_sequence():
    asn1_sequence = get_asn1_sequence(PRIVATE_KEY)
    assert (asn1_sequence == unhexlify(
        b'30818902818100d7ccfe871ad8cf4b2eee17d3a563b66679f4d529c14bb2ad26'
        b'484cd0ad6e704a236e61cf9da12576396c4f83ed5f55ed7d112e3701492afdb4'
        b'bf82300c2ceba73fabbd15b3c3a23ca1d55b3fe9356a7ed4b7b0f92543422dae'
        b'b67003086d1c2d16684cabc1824d392081ab7c29ba975086947258916a3309bd'
        b'fa20421b541c250203010001'
    ))


def test_calc_key_digest():
    key_digest = calc_key_digest(PRIVATE_KEY)
    assert hexlify(key_digest) == b'4e2a58768ccb6aa06f95e11646e187879d07fb66'


def test_calc_public_key_digest():
    public_key = PRIVATE_KEY.publickey()
    key_digest = calc_key_digest(public_key)
    assert hexlify(key_digest) == b'4e2a58768ccb6aa06f95e11646e187879d07fb66'


def test_calc_permanent_id():
    assert hexlify(calc_permanent_id(PRIVATE_KEY)) == b'4e2a58768ccb6aa06f95'


def test_calc_onion_address():
    assert calc_onion_address(PRIVATE_KEY) == u'jyvfq5umznvka34v'


def test_get_time_period():
    time_period = get_time_period(
        time=UNIX_TIMESTAMP,
        permanent_id=unhexlify(b'4e2a58768ccb6aa06f95'),
    )
    assert time_period == 16611


def test_get_seconds_valid():
    seconds_valid = get_seconds_valid(
        time=UNIX_TIMESTAMP,
        permanent_id=unhexlify(b'4e2a58768ccb6aa06f95'),
    )
    assert seconds_valid == 21054


def test_calc_secret_id_part():
    secret_id_part = calc_secret_id_part(
        time_period=16611,
        descriptor_cookie=None,
        replica=0,
    )
    assert (hexlify(secret_id_part) ==
            b'a0d8e4ec9ac28affed4fa828e8727c7fd4ab4930')


def test_calc_secret_id_part_descriptor_cookie():
    secret_id_part = calc_secret_id_part(
        time_period=16611,
        descriptor_cookie=base64.b64decode('dCmx3qIvArbil8A0KM4KgQ=='),
        replica=0,
    )
    assert (hexlify(secret_id_part) ==
            b'ea4e24b1a832f1da687f874b40fa9ecfe5221dd9')


def test_calc_descriptor_id():
    descriptor_id = calc_descriptor_id(
        permanent_id=b'N*Xv\x8c\xcbj\xa0o\x95',
        secret_id_part=unhexlify(b'a0d8e4ec9ac28affed4fa828e8727c7fd4ab4930'),
    )
    assert (hexlify(descriptor_id) ==
            b'f58ce3c63ee634e1ffaf936251a822ff06385f55')


def test_calc_descriptor_id_full():

    descriptor_id = calc_descriptor_id_b32(
        onion_address='jyvfq5umznvka34v',
        time=UNIX_TIMESTAMP,
        replica=0)

    assert descriptor_id == '6wgohrr64y2od75psnrfdkbc74ddqx2v'


def test_calc_descriptor_id_full_replica():
    descriptor_id = calc_descriptor_id_b32(
        onion_address='jyvfq5umznvka34v',
        time=UNIX_TIMESTAMP,
        replica=1)

    assert descriptor_id == 'he35m4nouhkz6thymvhdvc3y5htqs422'


def test_calc_descriptor_id_full_with_deviation():

    descriptor_id = calc_descriptor_id_b32(
        onion_address='jyvfq5umznvka34v',
        time=UNIX_TIMESTAMP,
        replica=0,
        deviation=1)

    assert descriptor_id == 'esnnz2q6dnfwprvc4qhsgsfzz6r6ksrt'


def test_rounded_timestamp():
    timestamp = datetime.datetime(2015, 6, 25, 13, 13, 25)
    assert rounded_timestamp(timestamp) == u'2015-06-25 13:00:00'


def test_rounded_timestamp_none_specified(monkeypatch):
    # Freeze datetime returned from datetime.datetime.utcnow()
    class frozen_datetime(datetime.datetime):
        @classmethod
        def utcnow(cls):
            return datetime.datetime(2015, 6, 25, 13, 13, 25)
    monkeypatch.setattr(datetime, 'datetime', frozen_datetime)
    assert rounded_timestamp(timestamp=None) == u'2015-06-25 13:00:00'


def test_base32_encode_str():
    assert base32_encode_str(byte_str=b'byte input') == u'mj4xizjanfxha5lu'


@pytest.mark.skipif(sys.version_info < (3, 0), reason="python3 only")
def test_base32_encode_str_not_byte_string():
    with pytest.raises(TypeError):
        base32_encode_str(byte_str=u'not a byte string')


def test_key_decrypt_prompt(mocker):
    # Valid private PEM key
    mocker.patch(builtin('open'), lambda *_: io.StringIO(PEM_PRIVATE_KEY))
    key = key_decrypt_prompt('private.key')
    assert isinstance(key, Cryptodome.PublicKey.RSA.RsaKey)
    assert key.has_private()


def test_key_decrypt_prompt_public_key(mocker):
    # Valid public PEM key
    private_key = Cryptodome.PublicKey.RSA.importKey(PEM_PRIVATE_KEY)
    pem_public_key = private_key.publickey().exportKey().decode('utf-8')
    mocker.patch(builtin('open'), lambda *_: io.StringIO(pem_public_key))

    with pytest.raises(ValueError):
        key_decrypt_prompt('public.key')


def test_key_decrypt_prompt_malformed_key(mocker):
    mocker.patch(builtin('open'), lambda *_: io.StringIO(PEM_INVALID_KEY))
    with pytest.raises(ValueError):
        key_decrypt_prompt('private.key')


def test_key_decrypt_prompt_incorrect_size(mocker):
    # Key which is not 1024 bits
    private_key_1280 = Cryptodome.PublicKey.RSA.generate(1280)
    pem_key_1280 = private_key_1280.exportKey().decode('utf-8')
    mocker.patch(builtin('open'), lambda *_: io.StringIO(pem_key_1280))
    with pytest.raises(ValueError):
        key_decrypt_prompt('512-bit-private.key')


def test_key_decrypt_prompt_encrypted(mocker):
    mocker.patch(builtin('open'), lambda *_: io.StringIO(PEM_ENCRYPTED))

    # Load with correct password
    mocker.patch('getpass.getpass', lambda *_: u'password')
    key = key_decrypt_prompt('encrypted_private.key')
    assert isinstance(key, Cryptodome.PublicKey.RSA.RsaKey)

    # Load with incorrect password
    mocker.patch('getpass.getpass', lambda *_: u'incorrect password')
    with pytest.raises(ValueError):
        key_decrypt_prompt('encrypted_private.key')


def test_try_make_dir_makedirs(mocker):
    mocker.patch('os.makedirs')
    try_make_dir('dir')
    os.makedirs.assert_called_once_with('dir')


def test_try_make_dir_makedirs_dir_already_exists(mocker):
    mocker.patch('os.makedirs', side_effect=OSError)
    mocker.patch('os.path.isdir', return_value=True)
    try_make_dir('dir')
    os.path.isdir.assert_called_once_with('dir')


def test_try_make_dir_makedirs_dir_other_error(mocker):
    mocker.patch('os.makedirs', side_effect=OSError)
    mocker.patch('os.path.isdir', return_value=False)
    with pytest.raises(OSError):
        try_make_dir('dir')


def test_is_directory_empty_empty(mocker):
    # Directory is empty
    mocker.patch('os.listdir', return_value=[])
    assert is_directory_empty('dir_empty/')


def test_is_directory_empty_not_empty(mocker):
    # Directory is empty
    mocker.patch('os.listdir', return_value=['filename'])
    assert not is_directory_empty('dir_not_empty/')



# -*- coding: utf-8 -*-
import base64
import os
import socket

import Cryptodome.PublicKey.RSA
import pytest
import yaml
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric.ed25519 import Ed25519PrivateKey
from cryptography.hazmat.primitives.serialization.base import Encoding, PublicFormat

from onionbalance.config_generator.config_generator import ConfigGenerator, parse_cmd_args

# Skip functional tests if Chutney environment is not running.
pytestmark = pytest.mark.skipif(
    "os.environ.get('CHUTNEY_ONION_ADDRESS') is None",
    reason="Skipping functional test, no Chutney environment detected")


def parse_chutney_enviroment():
    """
    Read environment variables and determine chutney instance and
    client addresses.
    """

    tor_client = os.environ.get('CHUTNEY_CLIENT_PORT')
    assert tor_client

    # Calculate the address and port of clients control port
    client_address, client_socks_port = tor_client.split(':')
    client_ip = socket.gethostbyname(client_address)

    tor_client_number = int(client_socks_port) - 9000
    # Control port in the 8000-8999 range, offset by Tor client number
    control_port = 8000 + tor_client_number
    assert control_port

    # Retrieve instance onion address exported during chutney setup
    instance_address = os.environ.get('CHUTNEY_ONION_ADDRESS')
    assert instance_address  # Need at least 1 instance address for test

    return {
        'client_ip': client_ip,
        'control_port': control_port,
        'instances': [instance_address],
    }


def create_test_config_file_v2(tmppath, private_key=None, instances=None):
    """
    Setup function to create a temp directory with master key and config file.
    Returns a path to the temporary config file.

    .. todo:: Refactor settings.py config creation to avoid code duplication
              in integration tests.
    """

    if not private_key:
        private_key = Cryptodome.PublicKey.RSA.generate(1024)

    # Write private key file
    key_path = tmppath.join('private_key')
    key_path.write(private_key.exportKey())
    assert key_path.check()

    # Create YAML Onionbalance settings file for these instances
    service_data = {'key': str(key_path)}
    service_data['instances'] = [{'address': addr[:16]} for addr in instances]
    settings_data = {
        'services': [service_data],
        'STATUS_SOCKET_LOCATION': str(tmppath.join('control')),
    }
    config_yaml = yaml.dump(settings_data, default_flow_style=False)

    config_path = tmppath.join('config.yaml')
    config_path.write_binary(config_yaml.encode('utf-8'))
    assert config_path.check()

    return str(config_path)


def create_test_config_file_v3(tmppath, instance_address):
    args = parse_cmd_args().parse_args(['--hs-version', 'v3', '-n', '1', '--output', str(tmppath)])
    ConfigGenerator(args, False)

    config_path = tmppath.join('config.yaml')
    assert config_path.check()

    with open(config_path) as f:
        config = yaml.safe_load(f)
    config['services'][0]['instances'][0]['address'] = instance_address

    with open(config_path, "w") as f:
        yaml.dump(config, f)

    return str(config_path)

def update_test_config_file_v3(tmppath, instance_address):
    config_path = tmppath.join('config.yaml')
    assert config_path.check()

    with open(config_path) as f:
        config = yaml.safe_load(f)
    config['services'][0]['instances'][0]['address'] = instance_address

    with open(config_path, "w") as f:
        yaml.dump(config, f)

    return str(config_path)

def random_onionv3_address():
    private_key = Ed25519PrivateKey.generate()
    public_key = private_key.public_key()
    public_bytes = public_key.public_bytes(encoding=Encoding.Raw, format=PublicFormat.Raw)

    # checksum = H(".onion checksum" || pubkey || version)
    checksumBytes = b''.join([b'.onion checksum', public_bytes, bytes([0x03])])
    digest = hashes.Hash(hashes.SHA3_256(), backend=default_backend())
    digest.update(checksumBytes)
    checksum = digest.finalize()[:2]

    # onion_address = base32(pubkey || checksum || version)
    onionAddressBytes = b''.join([public_bytes, checksum, bytes([0x03])])
    print(onionAddressBytes)
    onionAddress = base64.b32encode(onionAddressBytes).lower().decode('utf-8')

    return onionAddress + '.onion'



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""Convenience wrapper for running Onionbalance directly from source tree."""

import onionbalance.manager

if __name__ == '__main__':
    onionbalance.manager.main()



import datetime
import os

from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.backends import default_backend

import stem
import stem.descriptor.hidden_service
from stem.descriptor.hidden_service import HiddenServiceDescriptorV3

import onionbalance.common.descriptor
from onionbalance.common import log
import onionbalance.common.util

import onionbalance.hs_v3.instance
from onionbalance.hs_v3 import params
from onionbalance.hs_v3 import hashring
from onionbalance.hs_v3 import descriptor
from onionbalance.hs_v3 import tor_ed25519

logger = log.get_logger()


class OnionbalanceService(object):
    """
    Service represents a front-facing hidden service which should
    be load-balanced.
    """

    def __init__(self, service_config_data, config_path):
        """
        With 'config_data' straight out of the config file, create the service and its instances.
        'config_path' is the full path to the config file.

        Raise ValueError if the config file is not well formatted
        """
        # Is our private key in Tor's extended key format?
        self.is_priv_key_in_tor_format = False

        # Load private key and onion address from config
        # (the onion_address also includes the ".onion")
        self.identity_priv_key, self.onion_address = self._load_service_keys(service_config_data, config_path)

        # XXX This is an epic hack! If we are using keys in tor's extended
        # format, we basically override stem's function for signing with
        # blinded keys because it assumes that its keys are in standard
        # non-extended format. To avoid a double key extension we use our own
        # function...  This will prove to be a problem if we ever move to
        # multiple services per onionbalance, or if stem changes its code
        # behind our backs.
        if self.is_priv_key_in_tor_format:
            stem.descriptor.hidden_service._blinded_sign = tor_ed25519._blinded_sign_with_tor_key

        # Now load up the instances
        self.instances = self._load_instances(service_config_data)

        # First descriptor for this service (the one we uploaded last)
        self.first_descriptor = None
        # Second descriptor for this service (the one we uploaded last)
        self.second_descriptor = None

    def has_onion_address(self, onion_address):
        """
        Return True if this service has this onion address
        """
        # Strip the ".onion" part of the address if it exists since some
        # subsystems don't use it (e.g. Tor sometimes omits it from control
        # port responses)
        my_onion_address = self.onion_address.replace(".onion", "")
        their_onion_address = onion_address.replace(".onion", "")

        return my_onion_address == their_onion_address

    def _load_instances(self, service_config_data):
        instances = []

        for config_instance in service_config_data['instances']:
            new_instance = onionbalance.hs_v3.instance.InstanceV3(config_instance['address'])
            instances.append(new_instance)

        # Some basic validation
        for instance in instances:
            if self.has_onion_address(instance.onion_address):
                logger.error("Config file error. Did you configure your frontend (%s) as an instance?",
                             self.onion_address)
                raise BadServiceInit

        return instances

    def _load_service_keys(self, service_config_data, config_path):
        # First of all let's load up the private key
        key_fname = service_config_data['key']
        config_directory = os.path.dirname(config_path)
        if not os.path.isabs(key_fname):
            key_fname = os.path.join(config_directory, key_fname)

        try:
            with open(key_fname, 'rb') as handle:
                pem_key_bytes = handle.read()
        except EnvironmentError as e:
            logger.critical("Unable to read service private key file ('%s')", e)
            raise BadServiceInit

        # Get the service private key
        # First try with the OBv3 PEM format
        identity_priv_key = None
        try:
            identity_priv_key = serialization.load_pem_private_key(pem_key_bytes, password=None, backend=default_backend())
        except ValueError as e:
            logger.warning("Service private key not in OBv3 format ('%s'). Trying tor's format...", e)

        # If the key was not in OBv3 PEM format, try the Tor binary format
        if not identity_priv_key:
            try:
                identity_priv_key = tor_ed25519.load_tor_key_from_disk(pem_key_bytes)
                self.is_priv_key_in_tor_format = True
            except ValueError as e:
                logger.warning("Service private key not in Tor format either ('%s'). Aborting.", e)
                raise BadServiceInit

        # Get onion address
        identity_pub_key = identity_priv_key.public_key()
        identity_pub_key_bytes = identity_pub_key.public_bytes(encoding=serialization.Encoding.Raw,
                                                               format=serialization.PublicFormat.Raw)
        onion_address = HiddenServiceDescriptorV3.address_from_identity_key(identity_pub_key_bytes)

        logger.warning("Loaded onion %s from %s", onion_address, key_fname)

        return identity_priv_key, onion_address

    def _intro_set_modified(self, is_first_desc):
        """
        Check if the introduction point set has changed since last publish.
        """
        if is_first_desc:
            last_upload_ts = self.first_descriptor.last_upload_ts
        else:
            last_upload_ts = self.second_descriptor.last_upload_ts

        if not last_upload_ts:
            logger.info("\t Descriptor never published before. Do it now!")
            return True

        for instance in self.instances:
            if not instance.intro_set_modified_timestamp:
                logger.info("\t Still dont have a descriptor for this instance")
                continue

            if instance.intro_set_modified_timestamp > last_upload_ts:
                logger.info("\t Intro set modified")
                return True

        logger.info("\t Intro set not modified")
        return False

    def _get_descriptor_lifetime(self):
        from onionbalance.hs_v3.onionbalance import my_onionbalance
        if my_onionbalance.is_testnet:
            return params.FRONTEND_DESCRIPTOR_LIFETIME_TESTNET
        else:
            return params.FRONTEND_DESCRIPTOR_LIFETIME

    def _descriptor_has_expired(self, is_first_desc):
        """
        Check if the descriptor has expired (hasn't been uploaded recently).

        If 'is_first_desc' is set then check the first descriptor of the
        service, otherwise the second.
        """
        if is_first_desc:
            last_upload_ts = self.first_descriptor.last_upload_ts
        else:
            last_upload_ts = self.second_descriptor.last_upload_ts

        descriptor_age = (datetime.datetime.utcnow() - last_upload_ts)
        descriptor_age = int(descriptor_age.total_seconds())
        if (descriptor_age > self._get_descriptor_lifetime()):
            logger.info("\t Our %s descriptor has expired (%s seconds old). Uploading new one.",
                        "first" if is_first_desc else "second", descriptor_age)
            return True
        else:
            logger.info("\t Our %s descriptor is still fresh (%s seconds old).",
                        "first" if is_first_desc else "second", descriptor_age)
            return False

    def _hsdir_set_changed(self, is_first_desc):
        """
        Return True if the HSDir has changed between the last upload of this
        descriptor and the current state of things
        """
        from onionbalance.hs_v3.onionbalance import my_onionbalance

        # Derive blinding parameter
        _, time_period_number = hashring.get_srv_and_time_period(is_first_desc)
        blinded_param = my_onionbalance.consensus.get_blinding_param(self._get_identity_pubkey_bytes(),
                                                                     time_period_number)

        # Get blinded key
        # TODO: hoho! this is dirty we are poking into internal stem API. We
        #       should ask atagar to make it public for us! :)
        blinded_key = stem.descriptor.hidden_service._blinded_pubkey(self._get_identity_pubkey_bytes(), blinded_param)

        # Calculate current responsible HSDirs
        try:
            responsible_hsdirs = hashring.get_responsible_hsdirs(blinded_key, is_first_desc)
        except hashring.EmptyHashRing:
            return False

        if is_first_desc:
            previous_responsible_hsdirs = self.first_descriptor.responsible_hsdirs
        else:
            previous_responsible_hsdirs = self.second_descriptor.responsible_hsdirs

        if set(responsible_hsdirs) != set(previous_responsible_hsdirs):
            logger.info("\t HSDir set changed (%s vs %s)",
                        set(responsible_hsdirs), set(previous_responsible_hsdirs))
            return True
        else:
            logger.info("\t HSDir set remained the same")
            return False

    def _should_publish_descriptor_now(self, is_first_desc, force_publish=False):
        """
        Return True if we should publish a descriptor right now
        """
        # If descriptor not yet uploaded, do it now!
        if is_first_desc and not self.first_descriptor:
            return True
        if not is_first_desc and not self.second_descriptor:
            return True

        # OK this is not the first time we publish a descriptor. Check various
        # parameters to see if we should try to publish again:
        return any([self._intro_set_modified(is_first_desc),
                    self._descriptor_has_expired(is_first_desc),
                    self._hsdir_set_changed(is_first_desc),
                    force_publish])

    def get_all_intros_for_publish(self):
        """
        Return an IntroductionPointSetV3 with all the intros of all the instances
        of this service.
        """
        all_intros = []

        for instance in self.instances:
            try:
                instance_intros = instance.get_intros_for_publish()
            except onionbalance.hs_v3.instance.InstanceHasNoDescriptor:
                logger.info("Entirely missing a descriptor for instance %s. Continuing anyway if possible",
                            instance.onion_address)
                continue
            except onionbalance.hs_v3.instance.InstanceIsOffline:
                logger.info("Instance %s is offline. Ignoring its intro points...",
                            instance.onion_address)
                continue

            all_intros.append(instance_intros)

        return descriptor.IntroductionPointSetV3(all_intros)

    def publish_descriptors(self):
        self._publish_descriptor(is_first_desc=True)
        self._publish_descriptor(is_first_desc=False)

    def _get_intros_for_desc(self):
        """
        Get the intros that should be included in a descriptor for this service.
        """
        all_intros = self.get_all_intros_for_publish()

        # Get number of instances that contributed to final intro point list
        n_instances = len(all_intros.intro_points)
        n_intros_wanted = n_instances * params.N_INTROS_PER_INSTANCE

        final_intros = all_intros.choose(n_intros_wanted)

        if (len(final_intros) == 0):
            logger.info("Got no usable intro points from our instances. Delaying descriptor push...")
            raise NotEnoughIntros

        logger.info("We got %d intros from %d instances. We want %d intros ourselves (got: %d)",
                    len(all_intros.get_intro_points_flat()), n_instances,
                    n_intros_wanted, len(final_intros))

        return final_intros

    def _publish_descriptor(self, is_first_desc):
        """
        Attempt to publish descriptor if needed.

        If 'is_first_desc' is set then attempt to upload the first descriptor
        of the service, otherwise the second.
        """
        from onionbalance.hs_v3.onionbalance import my_onionbalance

        if not self._should_publish_descriptor_now(is_first_desc):
            logger.info("No reason to publish %s descriptor for %s",
                        "first" if is_first_desc else "second",
                        self.onion_address)
            return

        try:
            intro_points = self._get_intros_for_desc()
        except NotEnoughIntros:
            return

        # Derive blinding parameter
        _, time_period_number = hashring.get_srv_and_time_period(is_first_desc)
        blinding_param = my_onionbalance.consensus.get_blinding_param(self._get_identity_pubkey_bytes(),
                                                                      time_period_number)

        try:
            desc = descriptor.OBDescriptor(self.onion_address, self.identity_priv_key,
                                           blinding_param, intro_points, is_first_desc)
        except descriptor.BadDescriptor:
            return

        logger.info("Service %s created %s descriptor (%s intro points) (blinding param: %s) (size: %s bytes). About to publish:",
                    self.onion_address, "first" if is_first_desc else "second",
                    len(desc.intro_set), blinding_param.hex(), len(str(desc.v3_desc)))

        # When we do a v3 HSPOST on the control port, Tor decodes the
        # descriptor and extracts the blinded pubkey to be used when uploading
        # the descriptor. So let's do the same to compute the responsible
        # HSDirs:
        blinded_key = desc.get_blinded_key()

        # Calculate responsible HSDirs for our service
        try:
            responsible_hsdirs = hashring.get_responsible_hsdirs(blinded_key, is_first_desc)
        except hashring.EmptyHashRing:
            logger.warning("Can't publish desc with no hash ring. Delaying...")
            return

        desc.set_last_publish_attempt_ts(datetime.datetime.utcnow())

        logger.info("Uploading %s descriptor for %s to %s",
                    "first" if is_first_desc else "second",
                    self.onion_address, responsible_hsdirs)

        # Upload descriptor
        self._upload_descriptor(my_onionbalance.controller.controller,
                                desc, responsible_hsdirs)

        # It would be better to set last_upload_ts when an upload succeeds and
        # not when an upload is just attempted. Unfortunately the HS_DESC #
        # UPLOADED event does not provide information about the service and
        # so it can't be used to determine when descriptor upload succeeds
        desc.set_last_upload_ts(datetime.datetime.utcnow())
        desc.set_responsible_hsdirs(responsible_hsdirs)

        # Set the descriptor
        if is_first_desc:
            self.first_descriptor = desc
        else:
            self.second_descriptor = desc

    def _upload_descriptor(self, controller, ob_desc, hsdirs):
        """
        Convenience method to upload a descriptor
        Handle some error checking and logging inside the Service class
        """
        if hsdirs and not isinstance(hsdirs, list):
            hsdirs = [hsdirs]

        while True:
            try:
                onionbalance.common.descriptor.upload_descriptor(controller,
                                                                 ob_desc.v3_desc,
                                                                 hsdirs=hsdirs,
                                                                 v3_onion_address=ob_desc.onion_address)
                break
            except stem.SocketClosed:
                logger.error("Error uploading descriptor for service "
                             "%s.onion. Control port socket is closed.",
                             self.onion_address)
                onionbalance.common.util.reauthenticate(controller, logger)
            except stem.ControllerError:
                logger.exception("Error uploading descriptor for service "
                                 "%s.onion.", self.onion_address)
                break

    def _get_identity_pubkey_bytes(self):
        identity_pub_key = self.identity_priv_key.public_key()
        return identity_pub_key.public_bytes(encoding=serialization.Encoding.Raw,
                                             format=serialization.PublicFormat.Raw)


class NotEnoughIntros(Exception):
    pass

class BadServiceInit(Exception):
    pass



import unittest
import mock
import datetime
import base64

from cryptography.hazmat.primitives.asymmetric import ed25519

from onionbalance.hs_v3 import tor_node
from onionbalance.hs_v3 import hashring
from onionbalance.hs_v3 import consensus

CORRECT_HSDIR_FPRS_FIRST_DESCRIPTOR = [
    "D1D1D1D1D1D1D1D1D1D1D1D1D1D1D1D1D1D1D1D1",
    "2F2F2F2F2F2F2F2F2F2F2F2F2F2F2F2F2F2F2F2F",
    "B0B0B0B0B0B0B0B0B0B0B0B0B0B0B0B0B0B0B0B0",
    "3A3A3A3A3A3A3A3A3A3A3A3A3A3A3A3A3A3A3A3A",
    "5A5A5A5A5A5A5A5A5A5A5A5A5A5A5A5A5A5A5A5A",
    "DFDFDFDFDFDFDFDFDFDFDFDFDFDFDFDFDFDFDFDF",
    "F7F7F7F7F7F7F7F7F7F7F7F7F7F7F7F7F7F7F7F7",
    "3434343434343434343434343434343434343434" ]

CORRECT_HSDIR_FPRS_SECOND_DESCRIPTOR = [
    "5D5D5D5D5D5D5D5D5D5D5D5D5D5D5D5D5D5D5D5D",
    "9A9A9A9A9A9A9A9A9A9A9A9A9A9A9A9A9A9A9A9A",
    "D1D1D1D1D1D1D1D1D1D1D1D1D1D1D1D1D1D1D1D1",
    "7A7A7A7A7A7A7A7A7A7A7A7A7A7A7A7A7A7A7A7A",
    "C3C3C3C3C3C3C3C3C3C3C3C3C3C3C3C3C3C3C3C3",
    "C6C6C6C6C6C6C6C6C6C6C6C6C6C6C6C6C6C6C6C6",
    "E9E9E9E9E9E9E9E9E9E9E9E9E9E9E9E9E9E9E9E9",
    "8686868686868686868686868686868686868686"
]

class DummyConsensus(consensus.Consensus):
    def __init__(self):
        self.consensus = None


class TestHashRing(unittest.TestCase):
    def test_hashring(self):
        current_time = datetime.datetime.fromtimestamp(10101010101)
        current_srv = bytes([41])*32
        previous_srv = bytes([42])*32

        # Create 255 fake Tor nodes that will be used as part of the unittest
        network_nodes = []
        for i in range(1,256):
            microdescriptor = mock.Mock()
            routerstatus = mock.Mock()

            routerstatus.fingerprint = (bytes([i])*20).hex()
            routerstatus.protocols = {'HSDir' : [2]}
            routerstatus.flags = ['HSDir']
            node_ed25519_id_b64 = base64.b64encode(bytes([i])*32).decode('utf-8')
            microdescriptor.identifiers = {'ed25519' : node_ed25519_id_b64}
            node = tor_node.Node(microdescriptor, routerstatus)
            network_nodes.append(node)

        # Mock a fake consensus
        consensus = DummyConsensus()
        consensus.consensus = mock.Mock()
        consensus.consensus.valid_after = current_time
        consensus.get_current_srv = mock.Mock()
        consensus.get_current_srv.return_value = current_srv
        consensus.get_previous_srv = mock.Mock()
        consensus.get_previous_srv.return_value = previous_srv
        consensus.is_live = mock.Mock()
        consensus.is_live.return_value = True
        consensus.nodes = network_nodes

        # Mock a fake Tor network
        from onionbalance.hs_v3.onionbalance import my_onionbalance
        my_onionbalance.consensus = consensus

        previous_blinded_pubkey_hex = "063AEC5E1FD3025098F2DF71EF570B28D94B463FFCCB5EC6A9C061E38F551C6A"
        previous_blinded_pubkey_bytes = base64.b16decode(previous_blinded_pubkey_hex)

        responsible_hsdirs = hashring.get_responsible_hsdirs(previous_blinded_pubkey_bytes, True)

        i = 0
        for responsible_hsdir in responsible_hsdirs:
            self.assertEqual(responsible_hsdir.upper(), CORRECT_HSDIR_FPRS_FIRST_DESCRIPTOR[i])
            i+=1

        print("===")

        # we need to use the new blinded key since this uses a new time period.........
        current_blinded_pubkey_hex = "5DB624F2D74F103E6E8C6FBCCD074586EF5A5572F90673C00B77DEF94EC11499"
        current_blinded_pubkey_bytes = base64.b16decode(current_blinded_pubkey_hex)

        responsible_hsdirs = hashring.get_responsible_hsdirs(current_blinded_pubkey_bytes, False)

        i = 0
        for responsible_hsdir in responsible_hsdirs:
            self.assertEqual(responsible_hsdir.upper(), CORRECT_HSDIR_FPRS_SECOND_DESCRIPTOR[i])
            i+=1

if __name__ == '__main__':
    unittest.main()



import sys
import signal
import logging

from onionbalance.common import log
from onionbalance.hs_v3 import onionbalance as onionbalance_v3

logger = log.get_logger()


class SignalHandler(object):
    """
    Handle signals sent to the Onionbalance daemon process
    """

    def __init__(self, version, controller, status_socket=None):
        """
        Setup signal handler
        """
        self._version = version
        self._tor_controller = controller
        self._status_socket = status_socket

        # Register signal handlers
        signal.signal(signal.SIGTERM, self._handle_sigint_sigterm)
        signal.signal(signal.SIGINT, self._handle_sigint_sigterm)
        if self._version == 'v3':
            signal.signal(signal.SIGHUP, self._handle_sighup)

    def _handle_sigint_sigterm(self, signum, frame):
        """
        Handle SIGINT (Ctrl-C) and SIGTERM

        Disconnect from control port and cleanup the status socket
        """
        logger.info("Signal %d received, exiting", signum)
        self._tor_controller.close()
        if self._status_socket:
            self._status_socket.close()
        logging.shutdown()
        sys.exit(0)

    def _handle_sighup(self, signum, frame):
        """
        Handle SIGHUP (v3 only)

        Reload configuration
        """
        logger.info("Signal SIGHUP received, reloading configuration")
        onionbalance_v3.my_onionbalance.reload_config()



#! /usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Convenience wrapper for running onionbalance config generator directly from source tree.
"""

import onionbalance.config_generator.config_generator

if __name__ == '__main__':
    onionbalance.config_generator.config_generator.main()




# -*- coding: utf-8 -*-
import sys


def builtin(name):
    """
    Provide the correct import name for builtins on Python 2 or Python 3
    """
    if sys.version_info >= (3,):
        return 'builtins.{}'.format(name)
    else:
        return '__builtin__.{}'.format(name)



# -*- coding: utf-8 -*-
import base64
import textwrap
import datetime

import Cryptodome.Signature.pkcs1_15
import Cryptodome.Hash.SHA

import stem.descriptor.hidden_service_descriptor

from onionbalance.hs_v2 import util

from onionbalance.common import log
from onionbalance.hs_v2 import config

logger = log.get_logger()


def generate_service_descriptor(permanent_key, introduction_point_list=None,
                                replica=0, timestamp=None, deviation=0):
    """
    High-level interface for generating a signed HS descriptor
    """

    if not timestamp:
        timestamp = datetime.datetime.utcnow()
    unix_timestamp = int(timestamp.strftime("%s"))

    permanent_key_block = make_public_key_block(permanent_key)
    permanent_id = util.calc_permanent_id(permanent_key)

    # Calculate the current secret-id-part for this hidden service
    # Deviation allows the generation of a descriptor for a different time
    # period.
    time_period = (util.get_time_period(unix_timestamp, permanent_id) + int(deviation))

    secret_id_part = util.calc_secret_id_part(time_period, None, replica)
    descriptor_id = util.calc_descriptor_id(permanent_id, secret_id_part)

    if not introduction_point_list:
        onion_address = util.calc_onion_address(permanent_key)
        raise ValueError("No introduction points for service %s.onion." %
                         onion_address)

    # Generate the introduction point section of the descriptor
    intro_section = make_introduction_points_part(
        introduction_point_list
    )

    unsigned_descriptor = generate_hs_descriptor_raw(
        desc_id_base32=util.base32_encode_str(descriptor_id),
        permanent_key_block=permanent_key_block,
        secret_id_part_base32=util.base32_encode_str(secret_id_part),
        publication_time=util.rounded_timestamp(timestamp),
        introduction_points_part=intro_section
    )

    signed_descriptor = sign_descriptor(unsigned_descriptor, permanent_key)
    return signed_descriptor


def generate_hs_descriptor_raw(desc_id_base32, permanent_key_block,
                               secret_id_part_base32, publication_time,
                               introduction_points_part):
    """
    Generate hidden service descriptor string
    """
    doc = [
        "rendezvous-service-descriptor {}".format(desc_id_base32),
        "version 2",
        "permanent-key",
        permanent_key_block,
        "secret-id-part {}".format(secret_id_part_base32),
        "publication-time {}".format(publication_time),
        "protocol-versions 2,3",
        "introduction-points",
        introduction_points_part,
        "signature\n",
    ]

    unsigned_descriptor = '\n'.join(doc)
    return unsigned_descriptor


def make_introduction_points_part(introduction_point_list=None):
    """
    Make introduction point block from list of IntroductionPoint objects
    """

    # If no intro points were specified, we should create an empty list
    if not introduction_point_list:
        introduction_point_list = []

    intro = []
    for intro_point in introduction_point_list:
        intro.append("introduction-point {}".format(intro_point.identifier))
        intro.append("ip-address {}".format(intro_point.address))
        intro.append("onion-port {}".format(intro_point.port))
        intro.append("onion-key")
        intro.append(intro_point.onion_key)
        intro.append("service-key")
        intro.append(intro_point.service_key)

    intro_section = '\n'.join(intro).encode('utf-8')
    intro_section_base64 = base64.b64encode(intro_section).decode('utf-8')
    intro_section_base64 = textwrap.fill(intro_section_base64, 64)

    # Add the header and footer:
    intro_points_with_headers = '\n'.join([
        '-----BEGIN MESSAGE-----',
        intro_section_base64,
        '-----END MESSAGE-----'])
    return intro_points_with_headers


def make_public_key_block(key):
    """
    Get ASN.1 representation of public key, base64 and add headers
    """
    asn1_pub = util.get_asn1_sequence(key)
    pub_base64 = base64.b64encode(asn1_pub).decode('utf-8')
    pub_base64 = textwrap.fill(pub_base64, 64)

    # Add the header and footer:
    pub_with_headers = '\n'.join([
        '-----BEGIN RSA PUBLIC KEY-----',
        pub_base64,
        '-----END RSA PUBLIC KEY-----'])
    return pub_with_headers

def pad_msg_with_tor_pkcs(msg_hash, emLen, with_hash_parameters=True):
    """
    Tor requires PKCS#1 1.5 padding for descriptor signatures but does not
    include the algorithmIdentifier as specified in RFC3447.

    Unfortunately, most crypto libraries add that algorithmIdentifier by force
    and hence we need this function for monkey patching.
    """
    digestInfo = msg_hash.digest()
    PS = b'\xFF' * (emLen - len(digestInfo) - 3)
    return b'\x00\x01' + PS + b'\x00' + digestInfo

def sign(body, private_key):
    """
    Sign, base64 encode, wrap and add Tor signature headers

    The message digest is PKCS1 padded without the optional
    algorithmIdentifier section.
    """
    # First monkey-patch cryptodome to do the Tor PKCS#1 padding
    Cryptodome.Signature.pkcs1_15._EMSA_PKCS1_V1_5_ENCODE = pad_msg_with_tor_pkcs

    # The RSA signing API requires a hasher as input
    hasher = Cryptodome.Hash.SHA1.new(body)
    # Now do the signing
    signer = Cryptodome.Signature.pkcs1_15.new(private_key)
    signature_bytes = signer.sign(hasher)

    # Convert the signature to the base64 format that our descriptors like
    signature_base64 = base64.b64encode(signature_bytes).decode('utf-8')
    signature_base64 = textwrap.fill(signature_base64, 64)

    # Add the header and footer:
    signature_with_headers = '\n'.join([
        '-----BEGIN SIGNATURE-----',
        signature_base64,
        '-----END SIGNATURE-----'])
    return signature_with_headers


def sign_descriptor(descriptor, service_privkey):
    """
    Sign or resign a provided hidden service descriptor
    """
    token_descriptor_signature = '\nsignature\n'

    # Remove signature block if it exists
    if token_descriptor_signature in descriptor:
        descriptor = descriptor[:descriptor.find(token_descriptor_signature) + len(token_descriptor_signature)]
    else:
        descriptor = descriptor.strip() + token_descriptor_signature

    signature_with_headers = sign(descriptor.encode('utf-8'), service_privkey)
    return descriptor + signature_with_headers

def descriptor_received(descriptor_content):
    """
    Process onion service descriptors retrieved from the HSDir system or
    received directly over the metadata channel.
    """

    try:
        parsed_descriptor = stem.descriptor.hidden_service_descriptor.\
            HiddenServiceDescriptor(descriptor_content, validate=True)
    except ValueError:
        logger.exception("Received an invalid service descriptor.")
        return None

    # Ensure the received descriptor matches the requested descriptor
    permanent_key = Cryptodome.PublicKey.RSA.importKey(
        parsed_descriptor.permanent_key)
    descriptor_onion_address = util.calc_onion_address(permanent_key)

    known_descriptor, instance_changed = False, False
    for instance in [instance for service in config.services for
                     instance in service.instances]:
        if instance.onion_address == descriptor_onion_address:
            instance_changed |= instance.update_descriptor(parsed_descriptor)
            known_descriptor = True

    if instance_changed:
        logger.info("The introduction point set has changed for instance "
                    "%s.onion.", descriptor_onion_address)

    if not known_descriptor:
        # No matching service instance was found for the descriptor
        logger.debug("Received a descriptor for an unknown service:\n%s",
                     descriptor_content.decode('utf-8'))
        logger.warning("Received a descriptor with address %s.onion that "
                       "did not match any configured service instances.",
                       descriptor_onion_address)

    return None



# -*- coding: utf-8 -*-
"""
Base class to provide status over Unix socket
Default path: /var/run/onionbalance/control
"""
import errno
import os

from onionbalance.common import log


logger = log.get_logger()


class BaseStatusSocket(object):
    """
    For creating a Unix domain socket which emits a summary of the Onionbalance
    status when a client connects.
    """
    def __init__(self, unix_socket_filename):
        self.unix_socket_filename = unix_socket_filename

    def cleanup_socket_file(self):
        """
        Try to remove the socket file if it exists already
        """
        try:
            os.unlink(self.unix_socket_filename)
        except OSError as e:
            # Reraise if its not a FileNotFound exception
            if e.errno != errno.ENOENT:
                raise

    def close(self):
        """
        Close the unix domain socket and remove its file
        """
        try:
            self.server.shutdown()
            self.server.server_close()
            self.cleanup_socket_file()
        except AttributeError:
            pass
        except OSError:
            logger.exception("Error when removing the status socket")



# -*- coding: utf-8 -*-
import io
import os

import pytest

from onionbalance.hs_v2 import settings
from .util import builtin

CONFIG_FILE_VALID = u'\n'.join([
    "services:",
    "    - key: private.key",
    "      instances:",
    "        - address: fqyw6ojo2voercr7",
    "        - address: facebookcorewwwi",
])

CONFIG_FILE_ABSOLUTE = u'\n'.join([
    "services:",
    "    - key: /absdir/private.key",
    "      instances:",
    "        - address: fqyw6ojo2voercr7",
    "        - address: facebookcorewwwi",
])


def test_parse_config_file_valid(mocker):
    # Patch config file read
    mocker.patch('os.path.exists', return_value=True)
    mocker.patch(builtin('open'),
                 lambda *_: io.StringIO(CONFIG_FILE_VALID))

    parsed_config = settings.parse_config_file('/configdir/config_rel.yaml')

    assert len(parsed_config['services']) == 1
    assert len(parsed_config['services'][0]['instances']) == 2

    # Test key with absolute path
    assert os.path.dirname(parsed_config['services'][0]['key']) == '/configdir'

    # Test key with absolute path
    mocker.patch(builtin('open'),
                 lambda *_: io.StringIO(CONFIG_FILE_ABSOLUTE))
    parsed_config = settings.parse_config_file('/configdir/config_abs.yaml')
    assert os.path.dirname(parsed_config['services'][0]['key']) == '/absdir'


def test_parse_config_file_does_not_exist(mocker):
    with pytest.raises(SystemExit):
        settings.parse_config_file('doesnotexist/config.yaml')



import datetime
import unittest
import mock
from types import SimpleNamespace

from onionbalance.hs_v3 import consensus

from onionbalance.hs_v3.onionbalance import Onionbalance

class DummyConsensus(consensus.Consensus):
    def __init__(self):
        self.consensus = None

class OutdatedConsensus(unittest.TestCase):
    def test_outdated_consensus(self):
        current_time = datetime.datetime.fromtimestamp(10101010101)

        consensus = DummyConsensus()

        consensus.consensus = mock.Mock()
        with mock.patch("datetime.datetime") as mock_datetime:
            consensus.consensus.valid_after = current_time
            # valid_until is 3 hours in the future
            consensus.consensus.valid_until = current_time + datetime.timedelta(seconds=3600 * 3)

            # Test some legitimate cases
            mock_datetime.utcnow.return_value = current_time
            self.assertTrue(consensus.is_live())

            mock_datetime.utcnow.return_value = current_time + datetime.timedelta(seconds=3600 * 11)
            self.assertTrue(consensus.is_live())

            mock_datetime.utcnow.return_value = current_time + datetime.timedelta(seconds=3600 * 12)
            self.assertTrue(consensus.is_live())

            mock_datetime.utcnow.return_value = current_time + datetime.timedelta(seconds=3600 * 24)
            self.assertTrue(consensus.is_live())

            mock_datetime.utcnow.return_value = current_time - datetime.timedelta(seconds=3600 * 24)
            self.assertTrue(consensus.is_live())

            # Now test some bad cases. The is_live() function is lenient up to 24
            # hours after the valid_until, or 24 hours before the valid_after
            mock_datetime.utcnow.return_value = consensus.consensus.valid_until + datetime.timedelta(seconds=3600 * 24 + 1)
            self.assertFalse(consensus.is_live())

            mock_datetime.utcnow.return_value = consensus.consensus.valid_after - datetime.timedelta(seconds=3600 * 24 + 1)
            self.assertFalse(consensus.is_live())

class TestReloadConfig(unittest.TestCase):

    @mock.patch('onionbalance.hs_v3.service.OnionbalanceService')
    @mock.patch('onionbalance.hs_v3.consensus.Consensus')
    @mock.patch('onionbalance.hs_v3.stem_controller.StemController')
    @mock.patch('onionbalance.hs_v3.onionbalance.Onionbalance.load_config_file')
    @mock.patch('onionbalance.hs_v3.manager.init_scheduler')
    def test_reload_config(self, mock_init_scheduler, mock_load_config_file, mock_StemController, mock_Consensus, mock_OnionbalanceService):
        # setup onionbalance instance
        test_onionbalance = Onionbalance()
        test_onionbalance.args = self.create_dummy_args()

        # checks if Onionbalance.load_config_file and manager.init_scheduler have been called by reload_config
        test_onionbalance.reload_config()
        mock_load_config_file.assert_called_once()
        mock_init_scheduler.assert_called_once()

    @staticmethod
    def create_dummy_args():
        return SimpleNamespace(config='config/config.yaml', hs_version='v3', ip='127.0.0.1', is_testnet=False, port=6666, socket='/var/run/tor/control', verbosity='info')



import logging
import os
import sys

from onionbalance.common import scheduler
from onionbalance.common import log
from onionbalance.common import signalhandler

from onionbalance.hs_v3 import params
from onionbalance.hs_v3 import onionbalance

logger = log.get_logger()


def status_socket_location(config_data):
    location = os.environ.get('ONIONBALANCE_STATUS_SOCKET_LOCATION', '')
    if location == '':
        location = config_data.get('status-socket-location')
    return location


def main(args):
    """
    This is the entry point of v3 functionality.

    Initialize onionbalance, schedule future jobs and let the scheduler do its thing.
    """

    # Override the log level if specified on the command line.
    if args.verbosity:
        params.DEFAULT_LOG_LEVEL = args.verbosity.upper()
    logger.setLevel(logging.__dict__[params.DEFAULT_LOG_LEVEL.upper()])

    # Get the global onionbalance singleton
    my_onionbalance = onionbalance.my_onionbalance

    try:
        my_onionbalance.init_subsystems(args)
    except onionbalance.ConfigError as err:
        logger.error("%s", err)
        sys.exit(1)

    from onionbalance.hs_v3 import status
    status_socket = None
    if status_socket_location(my_onionbalance.config_data) is not None:
        status_socket = status.StatusSocket(status_socket_location(my_onionbalance.config_data), my_onionbalance)
    signalhandler.SignalHandler('v3', my_onionbalance.controller.controller, status_socket)

    # Schedule descriptor fetch and upload events
    init_scheduler(my_onionbalance)

    # Begin main loop to poll for HS descriptors
    scheduler.run_forever()

    return 0


def init_scheduler(my_onionbalance):
    scheduler.jobs = []

    if my_onionbalance.is_testnet:
        scheduler.add_job(params.FETCH_DESCRIPTOR_FREQUENCY_TESTNET, my_onionbalance.fetch_instance_descriptors)
        scheduler.add_job(params.PUBLISH_DESCRIPTOR_CHECK_FREQUENCY_TESTNET, my_onionbalance.publish_all_descriptors)
    else:
        scheduler.add_job(params.FETCH_DESCRIPTOR_FREQUENCY, my_onionbalance.fetch_instance_descriptors)
        scheduler.add_job(params.PUBLISH_DESCRIPTOR_CHECK_FREQUENCY, my_onionbalance.publish_all_descriptors)

    # Run initial fetch of HS instance descriptors
    scheduler.run_all(delay_seconds=params.INITIAL_CALLBACK_DELAY)
